var com;
(function (com) {
    var markenapps;
    (function (markenapps) {
        var thatlist;
        (function (thatlist) {
            var App = /** @class */ (function () {
                function App(data, assets) {
                    this.runningtime = 0;
                    this.tmp = {
                        catalog: [],
                        units: []
                    };
                    if (data.version !== window.appversion) {
                        log("update version from " + data.version);
                        this.version = window.appversion;
                    }
                    this.date = data.date || new Date().toJSON();
                    this.runningtime = data.runningtime || 0;
                    this.backupDate = data.backupDate;
                    //validate purchases
                    var validatedPurchases = [];
                    if (data.purchases) {
                        data.purchases.forEach(function (purchase) {
                            if (App.isProductValid(purchase)) {
                                validatedPurchases.push(purchase);
                            }
                        });
                    }
                    this.purchases = validatedPurchases;
                    var goods = assets.goods;
                    for (var g = 0; g < goods.length; g++) {
                        var curItem = goods[g];
                        this.tmp.catalog.push({
                            "cat": curItem[0],
                            "name": curItem[1],
                            "prio": curItem[2] || 0
                        });
                    }
                    if (data.categories) {
                        for (var catName in data.categories) {
                            var curAssetCatKeywords = (assets.categories[catName] && assets.categories[catName].keywords) || [];
                            //override with custom categories
                            var curCat = assets.categories[catName] = data.categories[catName];
                            curCat.keywords = curAssetCatKeywords.map(function (keyword) { return thatlist.util.normalizeStr(keyword); });
                        }
                    }
                    this.categories = assets.categories;
                    var units = assets.units;
                    for (var u = 0; u < units.length; u++) {
                        var curUnit = units[u];
                        //TODO do not normalize here
                        if (!curUnit.name)
                            curUnit.name = curUnit.title;
                        curUnit.normTitle = thatlist.util.normalizeStr(curUnit.title.split(" (")[0]);
                        if (curUnit.title_sing)
                            curUnit.normTitleSing = thatlist.util.normalizeStr(curUnit.title_sing.split(" (")[0]);
                    }
                    this.tmp.units = units;
                    this.settings = data.settings || {};
                    if (!this.settings.lang)
                        this.setLanguage(thatlist.util.getFirstBrowserLanguage());
                    if (!this.settings.fontSize)
                        this.settings.fontSize = 3;
                    if (this.settings.tips_hide === void (0))
                        this.settings.tips_hide = false;
                    var local_theme_name = localStorage.getItem("theme-name");
                    var local_theme_primaryColor = localStorage.getItem("theme-primaryColorIdx");
                    var local_theme_accentColor = localStorage.getItem("theme-accentColorIdx");
                    if (!this.settings.theme_name)
                        this.settings.theme_name = local_theme_name;
                    if (!this.settings.theme_primaryColor)
                        this.settings.theme_primaryColor = local_theme_primaryColor;
                    if (!this.settings.theme_accentColor)
                        this.settings.theme_accentColor = local_theme_accentColor;
                    if (local_theme_name !== this.settings.theme_name || local_theme_primaryColor !== !this.settings.theme_primaryColor || local_theme_accentColor !== !this.settings.theme_accentColor) {
                        //theme correction
                        setTheme(this.settings.theme_primaryColor, this.settings.theme_accentColor, this.settings.theme_name === "dark");
                    }
                    if (!data.lists) {
                        //disable for new installation by default
                        this.settings.lock_screen_feature = false;
                    }
                    else if (this.settings.lock_screen_feature === void (0)) {
                        this.settings.lock_screen_feature = true;
                    }
                    if (this.settings.force_portait)
                        thatlist.util.setOrientation('portrait');
                    if (this.settings.extract_count === void (0))
                        this.settings.extract_count = true;
                    this.lists = data.lists || [];
                    if (this.lists.length === 0) {
                        this.addList();
                    }
                    this.selectedListTS = data.selectedListTS || this.lists[0].ts;
                }
                App.getSupportedLanguages = function () {
                    return {
                        'da': 'Dansk',
                        'de': 'Deutsch',
                        'en': 'English',
                        'el': 'Ελληνικά',
                        'es': 'español',
                        'fr': 'français',
                        'it': 'italiano',
                        'nl': 'Nederlands',
                        'no': 'norsk',
                        'pl': 'polski',
                        'pt': 'português',
                        'fi': 'suomi',
                        'sv': 'svenska',
                        'cs': 'čeština',
                        'ru': 'русский',
                        'ja': '日本語'
                    };
                };
                App.isProductValid = function (product) {
                    var valid = allProductIds.indexOf(product.productId || product.id) >= 0;
                    if (!valid)
                        log("product invalid " + JSON.stringify(product));
                    return valid;
                };
                App.prototype.setLanguage = function (lang) {
                    this.settings.lang = thatlist.App.getSupportedLanguages()[lang] ? lang : '';
                };
                App.prototype.getLanguage = function () {
                    return app.settings.lang || thatlist.util.getFirstBrowserLanguage();
                };
                /**
                 * @returns true in case pruchases are loaded but there are no purchases
                 */
                App.prototype.hasAds = function () {
                    return this.purchases && !this.purchases.length;
                };
                App.prototype.onOwnedPurchase = function (handler) {
                    var _this = this;
                    if (!this.purchases)
                        this.purchases = [];
                    thatlist.util.onOwnedPurchase(allProductIds, function (ownedProduct) {
                        if (ownedProduct) {
                            log("ownedProduct " + JSON.stringify(ownedProduct));
                            if (!App.isProductValid(ownedProduct))
                                return;
                            _this.purchases = [ownedProduct];
                            app.save();
                            handler(_this.purchases);
                        }
                    });
                };
                App.prototype.buyProduct = function (productId) {
                    var _this = this;
                    return new Promise(function (resolve, reject) {
                        thatlist.util.buyProduct(productId).then(function (purchase) {
                            if (!purchase) {
                                reject();
                                return;
                            }
                            log("add purchase " + JSON.stringify(purchase));
                            _this.purchases = [purchase];
                            _this.save();
                            resolve(purchase);
                        }, function () {
                            reject();
                        });
                    });
                };
                App.prototype.addList = function () {
                    var catOrder = Object.keys(this.categories);
                    //filter custom categories title_iconname_color-hex-code
                    catOrder = catOrder.filter(function (catId) { return catId.length <= 10; });
                    var defOrder = ["FuV", "Ca", "SuD", "I", "D", "B", "S", "CuT", "C", "P", "Cu", "MuF", "Fish", "F", "R", "Sn", "DuH", "DIY", "O"];
                    catOrder.sort(function (id1, id2) {
                        var prio1 = defOrder.indexOf(id1);
                        if (prio1 < 0)
                            prio1 = defOrder.length;
                        var prio2 = defOrder.indexOf(id2);
                        if (prio2 < 0)
                            prio2 = defOrder.length;
                        if (prio1 > prio2)
                            return 1;
                        if (prio1 === prio2)
                            return 0;
                        return -1;
                    });
                    var newList = new List({ "catOrder": catOrder });
                    this.lists.push(newList);
                    return newList;
                };
                App.prototype.cloneList = function (list) {
                    var data = JSON.parse(JSON.stringify(list));
                    var newList = this.addList();
                    data.ts = Date.now();
                    data.title = list.title + "2";
                    for (var prop in data) {
                        console.log(prop);
                        newList[prop] = data[prop];
                    }
                    newList.disconnect();
                    return newList;
                };
                App.prototype.getList = function (ts) {
                    return this.lists.find(function (list) { return list.ts === ts; });
                };
                App.prototype.selectList = function (ts) {
                    log("Build new list!!!");
                    var list = new List(this.lists.find(function (list) { return list.ts == ts; }) || this.lists[0]);
                    this.selectedListTS = list.ts;
                    return list;
                };
                App.prototype.deleteList = function (list) {
                    var listIdx = this.indexOfList(list);
                    if (listIdx >= 0) {
                        this.lists.splice(listIdx, 1);
                    }
                };
                App.prototype.indexOfList = function (list) {
                    for (var l = 0; l < this.lists.length; l++) {
                        if (this.lists[l].ts === list.ts) {
                            return l;
                        }
                    }
                    return -1;
                };
                App.prototype.getFilteredUnits = function (parsedItem) {
                    if (!this.tmp.units || !this.tmp.units.length)
                        return [];
                    var filter = thatlist.util.normalizeStr(parsedItem.unit);
                    if (filter == "x")
                        filter = "";
                    var matchedUnits = this.tmp.units.filter(function (unit) {
                        var unitName = unit.normTitle;
                        var prio = 0;
                        if (unitName == filter) {
                            prio = 1000;
                        }
                        else {
                            var idx = unitName.indexOf(filter);
                            if (idx >= 0) {
                                prio = idx == 0 ? 800 : idx > 0 ? 600 : 0;
                                prio -= unitName.length;
                            }
                        }
                        if (unit.dP) {
                            prio += unit.dP + unitName.length;
                        }
                        if (!unit.tmp)
                            unit.tmp = {};
                        unit.tmp.prio = prio;
                        return prio;
                    }).sort(function (a, b) { return b.tmp.prio - a.tmp.prio; });
                    return matchedUnits;
                };
                App.prototype.hasPurchase = function (productId) {
                    for (var p = 0; p < this.purchases.length; p++) {
                        var curPurchase = this.purchases[p];
                        if (curPurchase.productId === productId) {
                            return true;
                        }
                    }
                    return false;
                };
                App.prototype.save = function (dataOrSave) {
                    var persistentData;
                    if (dataOrSave && dataOrSave !== true) {
                        log("save toJSON 1");
                        persistentData = typeof (dataOrSave) === "string" ? dataOrSave : thatlist.util.toJSON(dataOrSave);
                    }
                    else {
                        if (vm.list) {
                            var listIdx = this.indexOfList(vm.list);
                            if (listIdx >= 0) {
                                //replace ... otherwise primitive values will be lost because they are passed by copy and not be reference
                                //use splice ... otherwise vue will not recognize the change
                                this.lists.splice(listIdx, 1, vm.list);
                            }
                        }
                        log("save toJSON 2");
                        persistentData = thatlist.util.toJSON(this);
                    }
                    if (writeThrottle) {
                        clearTimeout(writeThrottle);
                        writeThrottle = null;
                    }
                    //if we save other data -> take care the we are not killed by next save
                    if (dataOrSave) {
                        thatlist.util.writeFile("shoppinglist.json", persistentData, true);
                        return Promise.resolve();
                    }
                    else {
                        return new Promise(function (resolve, reject) {
                            writeThrottle = setTimeout(function () {
                                thatlist.util.writeFile("shoppinglist.json", persistentData, true).then(resolve, reject);
                            }, 100);
                        });
                    }
                };
                App.prototype.hasCategory = function (catName) {
                    return !catName || this.categories[catName];
                };
                /**
                 * check if category needs to be created
                 * @param catName
                 *
                 * @returns true if category was created otherwise false
                 */
                App.prototype.checkCategoryName = function (catName) {
                    if (!this.hasCategory(catName)) {
                        //check if catName holds categorie properties
                        var catNameParts = catName.split("_");
                        if (catNameParts.length === 3) {
                            var catTitle = catNameParts[0];
                            var catIconName = "icon-" + catNameParts[1];
                            var catColor = "#" + catNameParts[2];
                            if (catColor.length === 7 && iconNames.indexOf(catIconName) >= 0) {
                                var newCat = {
                                    "name": catName,
                                    "title": catTitle,
                                    "icon": catIconName,
                                    "color": catColor,
                                    "persist": true
                                };
                                this.categories[catName] = newCat;
                                return true;
                            }
                        }
                    }
                    return false;
                };
                App.prototype.rate = function () {
                    log("TODO check rate due");
                };
                App.prototype.doBackup = function (force) {
                    var _this = this;
                    console.log(force, this.backupDate);
                    if (force || !this.backupDate || Date.now() - new Date(this.backupDate).getTime() > 1000 * 60 * 60 * 12) {
                        console.log("a");
                        return new Promise(function (resolve, reject) {
                            try {
                                var purchasesBck = _this.purchases;
                                delete _this.purchases;
                                var data = thatlist.util.toJSON(_this);
                                data = data.replace(/\"commands\":/g, "\"_commands_\":");
                                data = JSON.parse(data);
                                data.backupDate = new Date().toJSON();
                                _this.purchases = purchasesBck;
                                var zip = new JSZip();
                                //add year and month to delete outdated backups on backend
                                console.log(data);
                                zip.file("_tsl_" + (data.backupDate.substr(0, 7)) + ".json", thatlist.util.toJSON(data));
                                zip.generateAsync({
                                    type: "blob",
                                    compression: "DEFLATE",
                                    compressionOptions: {
                                        level: 9
                                    }
                                }).then(function (content) {
                                    thatlist.util.requestBlob(getBackupPath(), "POST", content).then(function () {
                                        _this.backupDate = data.backupDate;
                                        _this.save();
                                        resolve();
                                    }, reject);
                                }, reject);
                            }
                            catch (e) {
                                log(e);
                                reject();
                            }
                        });
                    }
                    else {
                        return Promise.resolve();
                    }
                };
                App.prototype.loadBackup = function (backupId, noWarn) {
                    var _this = this;
                    return new Promise(function (resolve, reject) {
                        thatlist.util.requestBlob(getBackupPath(backupId))
                            .then(JSZip.loadAsync, function (status) {
                            if (status === 404) {
                                _this.backupDate = void (0);
                                reject();
                            }
                        })
                            .then(function (zip) {
                            return zip.file(Object.keys(zip.files)[0]).async("string");
                        }, reject)
                            .then(function (text) {
                            try {
                                var data = JSON.parse(text);
                                resolve(data);
                            }
                            catch (e) {
                                reject();
                            }
                        }, reject);
                    });
                };
                App.prototype.deleteBackup = function (backupId) {
                    var _this = this;
                    return new Promise(function (resolve, reject) {
                        var bckDelPromise = thatlist.util.request(getBackupPath(backupId), "DELETE").then(function () {
                            _this.backupDate = void (0);
                            resolve();
                        }, reject);
                    });
                };
                return App;
            }());
            thatlist.App = App;
            var List = /** @class */ (function () {
                function List(data) {
                    var _this = this;
                    this.pickedItemsLimit = 30;
                    this.tmp = {};
                    this.sortItemsByCategory = function (a, b) {
                        var pos = _this.catOrder.indexOf(a.cat) - _this.catOrder.indexOf(b.cat);
                        if (pos != 0)
                            return pos;
                        return thatlist.util.normalizeStr(a.name) < thatlist.util.normalizeStr(b.name) ? -1 : 1;
                    };
                    this.ts = data.ts || Date.now();
                    this.title = data.title || $translate.instant("supermarket");
                    //migrate do not persist
                    //this.itemNames = data.itemNames || {};
                    this.items = data.items || []; //(data.items || []).map(item => {item.tmp = {collapsed: false}; return item;});
                    // this.items.forEach(item => {if (item && item.name) this.tmp.itemsByNames[item.name] = item});
                    this.pickedItems = data.pickedItems || [];
                    this.pickedItems = this.pickedItems.splice(0, 30);
                    if (data.pickedItemsLimit !== void (0))
                        this.pickedItemsLimit = data.pickedItemsLimit;
                    //clear inconsistent catOrder for dublicates and null
                    var clearedCatOrder;
                    if (data.catOrder) {
                        clearedCatOrder = [];
                        var cats = {};
                        data.catOrder.forEach(function (cat) {
                            if (!cat)
                                return;
                            if (!cats[cat]) {
                                cats[cat] = true;
                                clearedCatOrder.push(cat);
                            }
                        });
                    }
                    this.catOrder = clearedCatOrder;
                    this.id = data.id || "";
                    this.token = data.token || "";
                    this.remoteVersion = data.remoteVersion;
                    this.commands = data.commands || [];
                    var favsByName = {};
                    if (data.favourites) {
                        if (!Array.isArray(data.favourites)) {
                            var favourites = [];
                            //migrate to array
                            for (var name in data.favourites) {
                                var curFav = data.favourites[name];
                                if (!curFav || !name)
                                    continue;
                                //migrate use [data, interval, use] -> prio
                                favourites.push({
                                    "name": name,
                                    "cat": curFav.cat,
                                    "prio": curFav.prio || curFav.use[2] || 1
                                });
                            }
                            data.favourites = favourites;
                        }
                        var doDowngrade_1 = false;
                        var recentDowngradeTS = parseInt(localStorage.getItem("recentDowngradeTS") || "0", 10);
                        //downgrade prio once a day only
                        if (!recentDowngradeTS || Date.now() - recentDowngradeTS > 60000 * 60 * 24) {
                            localStorage.setItem("recentDowngradeTS", Date.now());
                            doDowngrade_1 = true;
                        }
                        data.favourites.forEach(function (fav) {
                            //fix invisible favs
                            if (!fav.prio)
                                fav.prio = 1;
                            favsByName[fav.name] = fav;
                            //naive downgrade prio/use count 
                            if (doDowngrade_1 && fav.prio > 4) {
                                //fav with prio 4 is downgraded by 1 when it's not added on 50 usage days
                                fav.prio = Math.max(4, fav.prio - (fav.prio / 200));
                                //round for two decimals
                                fav.prio = Math.round(fav.prio * 100) / 100;
                            }
                        });
                    }
                    this.favourites = data.favourites || [];
                    //because we init fist list in constuctor of app
                    if (app && app.tmp && app.tmp.catalog) {
                        var catalog = JSON.parse(JSON.stringify(app.tmp.catalog)).filter(function (item) { return !favsByName[item.name]; });
                        this.tmp.catalog = catalog.concat(this.favourites);
                    }
                }
                List.prototype.addCommand = function (command) {
                    if (command.cmd && command.picked)
                        throw "command can have either a cmd or a picked property";
                    if (command && command.cmd && command.good) {
                        //handle conflicts
                        for (var c = this.commands.length; c--;) {
                            var curCmd = this.commands[c];
                            if (command.good.name === curCmd.good.name) {
                                //dublicate
                                if (curCmd.cmd === command.cmd) {
                                    //update old with new 
                                    this.commands[c] = command;
                                    return;
                                }
                                else if (curCmd.cmd === "add" && command.cmd === "del") {
                                    // TODO ignore both when add is unsynced
                                    this.commands.splice(c, 1);
                                    break;
                                }
                                else if (curCmd.cmd === "del" && command.cmd === "add") {
                                    //ignore del and apply add (may be count differs)
                                    this.commands.splice(c, 1);
                                    break;
                                }
                            }
                        }
                        this.commands.push({
                            "cmd": command.cmd,
                            "good": {
                                "cat": command.good.cat,
                                "name": command.good.name,
                                "count": command.good.count
                            }
                        });
                    }
                    if (command && command.picked && command.good) {
                        //handle conflicts
                        for (var c = this.commands.length; c--;) {
                            var curCmd = this.commands[c];
                            //backward compatibility
                            if (curCmd.cmd)
                                continue;
                            if (command.good.name === curCmd.good.name) {
                                //dublicate
                                if (curCmd.picked === command.picked) {
                                    //update old with new 
                                    this.commands[c] = command;
                                    return;
                                }
                                else if (curCmd.picked === "add" && command.picked === "del") {
                                    //ignore add and apply del TODO ignore both when add is unsynced
                                    this.commands.splice(c, 1);
                                    break;
                                }
                                else if (curCmd.picked === "del" && command.picked === "add") {
                                    //ignore del and apply add (may be count differs)
                                    this.commands.splice(c, 1);
                                    break;
                                }
                            }
                        }
                        this.commands.push({
                            "picked": command.picked,
                            "name": command.name,
                            "good": {
                                "cat": command.good.cat,
                                "name": command.good.name,
                                "count": command.good.count
                            }
                        });
                    }
                    log(thatlist.util.toJSON(this.commands));
                };
                List.prototype.applyCommands = function (options) {
                    var _this = this;
                    this.commands.forEach(function (command) {
                        var item = command.good;
                        if (!item || !item.name)
                            return;
                        if (command.cmd === "add") {
                            _this.checkCategory(item);
                            log("add ", item);
                            _this.addItem(item, options);
                        }
                        else if (command.cmd === "del") {
                            var curItemIdx = thatlist.util.indexOfItem(item.name, _this.items);
                            if (curItemIdx >= 0) {
                                _this.items.splice(curItemIdx, 1);
                            }
                        }
                        else if (command.cmd === "update") {
                            //if not there do not insert ... may be it was deleted
                            var idx = thatlist.util.indexOfItem(item.name, _this.items);
                            if (idx >= 0) {
                                _this.items.splice(idx, 1, item);
                            }
                            else {
                                return;
                            }
                        }
                    });
                };
                List.prototype.getFavouriteByName = function (itemName) {
                    return this.favourites[thatlist.util.indexOfItem(itemName, this.favourites)];
                };
                List.prototype.getItemByName = function (itemName) {
                    return this.items[thatlist.util.indexOfItem(itemName, this.items)];
                };
                List.prototype.hasCategory = function (catName) {
                    return this.catOrder.indexOf(catName) >= 0;
                };
                List.prototype.hideCategory = function (category) {
                    var idx = this.catOrder.indexOf(category.name);
                    if (idx >= 0)
                        this.catOrder.splice(idx, 1);
                };
                List.prototype.unhideCategory = function (category) {
                    var idx = this.catOrder.indexOf(category.name);
                    if (idx < 0)
                        this.catOrder.push(category.name);
                };
                /**
                 * Detect category if non is present and add category if missing or hidden
                 * @param item
                 */
                List.prototype.checkCategory = function (item) {
                    var origCat = item.cat;
                    //detailed check category (based on catalog)
                    if (!item.cat) {
                        var g = void 0;
                        try {
                            var bestCatRank = 100;
                            for (g = 0; g < this.tmp.catalog.length; g++) {
                                var curItem = this.tmp.catalog[g];
                                //check if cur good categorie is a visible categorie
                                if (!this.hasCategory(curItem.cat))
                                    continue;
                                ////remove plural
                                var itemSearchName = thatlist.util.normalizeStr(item.name); //.length>=4 ? good.normName.substring(0, good.normName.length-2) : good.normName;
                                var curItemSearchName = thatlist.util.normalizeStr(curItem.name); //.length>=4 ? curGood.normName.substring(0, curGood.normName.length-2) : curGood.normName;
                                if (!curItemSearchName) {
                                    continue;
                                }
                                var curCatRank = Math.abs(curItemSearchName.length - itemSearchName.length);
                                if (curCatRank > bestCatRank)
                                    continue;
                                if (curItem && curItemSearchName && curItem.cat && curItem.cat !== "O") {
                                    //avoid öl (ol) match rucola but let ei match eier
                                    if (curItemSearchName.indexOf(itemSearchName) >= 0) {
                                        log(".cat= 3");
                                        item.cat = curItem.cat;
                                        bestCatRank = curCatRank;
                                        if (curCatRank === 0)
                                            break;
                                    }
                                    else if (itemSearchName.indexOf(curItemSearchName) >= 0) {
                                        log(".cat= 4");
                                        item.cat = curItem.cat;
                                        bestCatRank = curCatRank;
                                        if (curCatRank === 0)
                                            break;
                                    }
                                }
                            }
                        }
                        catch (e) {
                            window.logError("APP ERROR get cat: " + e.message + (e.stack ? "\n" + e.stack : "") + "-" + this.tmp.catalog + "-" + item.name + "-" + this.tmp.catalog[g].name);
                        }
                    }
                    //fallback detect category (based on keywords)
                    if (!item.cat) { // || item.cat === "O") {
                        delete item.cat;
                        //detect cat
                        var bestCat_1;
                        var bestCatPrio_1 = -1;
                        this.catOrder.forEach(function (catName) {
                            var curCatPrio = 0;
                            var filter = thatlist.util.normalizeStr(item.name);
                            var curCat = app.categories[catName];
                            if (curCat && filter) {
                                for (var k = 0; curCat.keywords && k < curCat.keywords.length; k++) {
                                    var curKeyword = curCat.keywords[k];
                                    if (filter.indexOf(curKeyword) >= 0) {
                                        curCatPrio = curKeyword.length;
                                        break;
                                    }
                                }
                            }
                            if (curCatPrio && curCatPrio >= bestCatPrio_1) {
                                bestCat_1 = curCat;
                                bestCatPrio_1 = curCatPrio;
                            }
                        });
                        if (bestCat_1) {
                            log(".cat= 2");
                            item.cat = bestCat_1.name;
                        }
                    }
                    if (!item.cat) {
                        log(".cat= 5");
                        item.cat = origCat || "O";
                    }
                    app.checkCategoryName(item.cat);
                    if (item.cat && !this.hasCategory(item.cat))
                        this.catOrder.push(item.cat);
                };
                List.prototype.addItem = function (item, options) {
                    if (!item || !item.name) {
                        return;
                    }
                    if (!options)
                        options = {};
                    if (!item.name)
                        return;
                    item.name = item.name.trim();
                    this.checkCategory(item);
                    if (item.tmp && item.tmp.count) {
                        item.count = item.tmp.count;
                        delete item.tmp.count;
                    }
                    // check if this item is already in list
                    var curItemIdx = thatlist.util.indexOfItem(item.name, this.items);
                    // add/update command
                    this.addCommand({
                        "cmd": "add",
                        "good": item,
                    });
                    if (curItemIdx >= 0) {
                        //do update prio when not already in items list
                        if (typeof options.keepPrio === void (0))
                            options.keepPrio = true;
                        //replace item
                        this.items.splice(curItemIdx, 1, item)[0];
                    }
                    else {
                        if (!options.bottom)
                            this.items.unshift(item);
                        else
                            this.items.push(item);
                    }
                    //add/update fav (may be cat or count was changed)
                    this.addFavourite(item, options);
                    if (options.sort || options.sort === void (0))
                        this.sort();
                };
                List.prototype.addItems = function (items, options) {
                    var _this = this;
                    if (options === void 0) { options = {}; }
                    items.forEach(function (item) {
                        //add item only if not in list already
                        if (!_this.getItemByName(item.name)) {
                            _this.addItem(item, { "sort": false });
                        }
                    });
                    if (options.clearPicked) {
                        this.pickedItems = [];
                    }
                    this.sort();
                };
                List.prototype.parseItemsFromSpeech = function (text) {
                    var units = app.tmp.units;
                    var result = text;
                    var items = result.split(thatlist.util.andSplitWords[app.getLanguage()]);
                    var itemsCollection = [];
                    items.forEach(function (item) {
                        var cleanItemText = item.trim();
                        if (!cleanItemText)
                            return;
                        var itemParts = cleanItemText.split(" ");
                        var precWasNumber;
                        for (var p = 0; p < itemParts.length; p++) {
                            var part = itemParts[p];
                            var normPart = part.toLowerCase();
                            var isUnit = false;
                            var num = isNaN(part) ? thatlist.util.numberWordToNumber(normPart) : part;
                            if (num) {
                                itemParts[p] = num;
                                precWasNumber = true;
                            }
                            else if (precWasNumber) {
                                precWasNumber = false;
                                //check if unit
                                for (var u = 0; u < units.length; u++) {
                                    var curUnit = units[u];
                                    //curUnit.normTitle = curUnit.normTitle === void (0) ? curUnit.title.split(" (")[0].toLowerCase() : curUnit.normTitle;
                                    //curUnit.normTitleSing = curUnit.title_sing && curUnit.normTitleSing === void (0) ? curUnit.title_sing.split(" (")[0].toLowerCase() : curUnit.normTitleSing;
                                    if (normPart === curUnit.normTitle)
                                        isUnit = true;
                                    else if (normPart === curUnit.normTitleSing)
                                        isUnit = true;
                                    if (isUnit) {
                                        //replace long unit with short unit name
                                        if (curUnit.name)
                                            itemParts[p] = part = curUnit.name;
                                        break;
                                    }
                                }
                                //heuristik for unknown units
                                if (!isUnit && itemParts.length > 2 && p === itemParts.length - 1) {
                                    //guess in case of "name count unit" e.g. Milch mit wenig Fett 3 Liter
                                    isUnit = true;
                                }
                                //join amount and unit
                                if (isUnit) {
                                    //push amount and unit to the start
                                    itemParts.unshift(itemParts[p - 1] + part);
                                    //remove from the end
                                    itemParts.splice(p, 2);
                                }
                            }
                        }
                        var item = thatlist.util.parseItemName(itemParts.join(" "));
                        if (item) {
                            // if (item.name && item.name.length > 1) item.selected = true;
                            itemsCollection.push(item);
                        }
                    });
                    return itemsCollection;
                };
                List.prototype.removeItem = function (item) {
                    var idx = thatlist.util.indexOfItem(item.name, this.items);
                    if (idx >= 0) {
                        this.addCommand({
                            "cmd": "del",
                            "good": item
                        });
                        return this.items.splice(idx, 1)[0];
                    }
                    else {
                        return null;
                    }
                };
                List.prototype.clearItems = function () {
                    var _this = this;
                    if (this.token) {
                        this.items.forEach(function (item) {
                            _this.addCommand({
                                "cmd": "del",
                                "good": item
                            });
                        });
                    }
                    else {
                        //clear commands if list is not synced
                        this.commands.length = 0;
                    }
                    this.items = [];
                };
                List.prototype.addPickedItem = function (item, options) {
                    if (options === void 0) { options = {}; }
                    if (!item || !item.name)
                        return;
                    if (!app.settings.keep_count) {
                        if (!item.tmp)
                            item.tmp = {};
                        item.tmp.pickedCount = item.count;
                        var curFav = this.getFavouriteByName(item.name);
                        if (curFav)
                            delete curFav.count;
                        delete item.count;
                    }
                    var insertAt = 0;
                    if (options.replaceItem) {
                        var idx = thatlist.util.indexOfItem(options.replaceItem.name, this.pickedItems);
                        if (idx >= 0) {
                            insertAt = idx;
                            this.addCommand({ "picked": "replace", "name": options.replaceItem.name, "good": item });
                            this.pickedItems.splice(idx, 1);
                        }
                    }
                    else {
                        var idx = thatlist.util.indexOfItem(item.name, this.pickedItems);
                        if (idx >= 0)
                            this.pickedItems.splice(idx, 1);
                        this.addCommand({ "picked": "add", "good": item });
                        this.pickedItemsLimit++;
                    }
                    this.pickedItems.splice(insertAt, 0, item);
                };
                List.prototype.removePickedItem = function (item) {
                    var curPickedItemIdx = thatlist.util.indexOfItem(item.name, this.pickedItems);
                    if (curPickedItemIdx >= 0) {
                        if (this.pickedItemsLimit && curPickedItemIdx < this.pickedItemsLimit)
                            this.pickedItemsLimit--;
                        this.addCommand({
                            "picked": "del",
                            "good": item
                        });
                        return this.pickedItems.splice(curPickedItemIdx, 1)[0];
                    }
                };
                List.prototype.hidePickedItems = function () {
                    this.pickedItemsLimit = 0;
                };
                /**
                 * add/update fav if not already present
                 * @param item item to add
                 */
                List.prototype.addFavourite = function (item, options) {
                    if (!item || !item.name)
                        return;
                    if (!options)
                        options = {};
                    try {
                        //check fav
                        var curFav = this.getFavouriteByName(item.name);
                        if (!curFav) {
                            //add favourite
                            curFav = JSON.parse(JSON.stringify(item));
                            this.favourites.unshift(curFav);
                            //update catalog
                            var curItemIdx = thatlist.util.indexOfItem(item.name, this.tmp.catalog);
                            if (curItemIdx < 0) {
                                this.tmp.catalog.unshift(curFav);
                            }
                            else {
                                //reuse cur prio of good
                                if (!curFav.prio)
                                    curFav.prio = this.tmp.catalog[curItemIdx].prio;
                                this.tmp.catalog.splice(curItemIdx, 1, curFav);
                            }
                        }
                        else {
                            log(".cat= 6", !!curFav, !!item);
                            curFav.cat = item.cat;
                            curFav.count = item.count;
                        }
                        if (!app.settings.keep_count) {
                            delete curFav.count;
                        }
                        var curUsedDate = new Date().toJSON().substring(0, 10);
                        if (!curFav.tmp)
                            curFav.tmp = {};
                        var recentUsedDate = curFav.tmp.usedDate;
                        if (!curFav.prio) {
                            curFav.prio = 3;
                        }
                        else {
                            //avoid double use by corretion: done->add from history->done again
                            if (!options.keepPrio && recentUsedDate !== curUsedDate) {
                                curFav.tmp.usedDate = curUsedDate;
                                if (curFav.prio < 10)
                                    curFav.prio++;
                            }
                        }
                    }
                    catch (e) {
                        window.logError("APP ERROR update fav: " + e.message + (e.stack ? "\n" + e.stack : ""));
                    }
                };
                /**
                 * remove item if present
                 * @param item item to remove
                 */
                List.prototype.removeFavourite = function (item) {
                    if (!item || !item.name)
                        return;
                    var idx = thatlist.util.indexOfItem(item.name, this.favourites);
                    if (idx >= 0) {
                        //hide from search result until restart
                        item.prio = 0;
                        this.favourites.splice(idx, 1);
                    }
                };
                List.prototype.deleteItem = function (item) {
                    if (!item || !item.name)
                        return;
                    var idx = thatlist.util.indexOfItem(item.name, this.tmp.catalog);
                    if (idx >= 0) {
                        this.tmp.catalog.splice(idx, 1);
                    }
                    this.removeItem(item);
                    this.removePickedItem(item);
                    this.removeFavourite(item);
                };
                List.prototype.sort = function () {
                    this.items.sort(this.sortItemsByCategory);
                    app.save();
                };
                List.prototype.getCatalogItems = function (item) {
                    var count = item.count;
                    var filter = thatlist.util.normalizeStr(item.name);
                    if (!filter)
                        return [];
                    var catsByName = {};
                    this.catOrder.forEach(function (catName) { catsByName[catName] = true; });
                    var items = this.tmp.catalog.filter(function (item) {
                        if (!catsByName[item.cat] || item.prio === 0)
                            return false;
                        //var goodCat = good.normCat || normalizeStr(good.cat);
                        var itemName = thatlist.util.normalizeStr(item.name);
                        var prio = 0;
                        if (itemName == filter) {
                            prio = 1000;
                        }
                        else {
                            var idx = itemName.indexOf(filter);
                            if (idx >= 0) {
                                prio = idx == 0 ? 800 : idx > 0 ? 600 : 0;
                                if (idx > 0) {
                                    prio -= itemName.length;
                                }
                            }
                            if (item.name === "Mouka")
                                console.log("Mouka1 ", prio);
                            if (item.name === "Mléko")
                                console.log("Mléko1 ", prio);
                        }
                        if (prio > 0 && item.prio) {
                            prio += item.prio;
                        }
                        if (item.name === "Mouka")
                            console.log("Mouka2 ", prio);
                        if (item.name === "Mléko")
                            console.log("Mléko2 ", prio);
                        if (!item.tmp)
                            item.tmp = {};
                        item.tmp.count = count;
                        item.tmp.prio = prio;
                        return prio;
                    }).sort(function (a, b) { return b.tmp.prio - a.tmp.prio; }).slice(0, 8);
                    return items;
                };
                List.prototype.getFavourites = function (orderBy, limit) {
                    var _this = this;
                    if (limit === 0)
                        return [];
                    log("calc favorites by " + orderBy + " " + limit);
                    var favourites = this.favourites.filter(function (item) {
                        return thatlist.util.indexOfItem(item.name, _this.items) < 0;
                    });
                    favourites.sort(function (a, b) { return b.prio - a.prio; });
                    if (orderBy === "cat") {
                        favourites.sort(function (a, b) {
                            var asv = "0000" + _this.catOrder.indexOf(a.cat);
                            asv = asv.substring(asv.length - 4) + thatlist.util.normalizeStr(a.name);
                            var bsv = "0000" + _this.catOrder.indexOf(b.cat);
                            bsv = bsv.substring(bsv.length - 4) + thatlist.util.normalizeStr(b.name);
                            if (asv > bsv)
                                return 1;
                            if (asv < bsv)
                                return -1;
                            return 0;
                        });
                    }
                    if (limit)
                        favourites = favourites.splice(0, limit);
                    return favourites;
                };
                List.prototype.connect = function () {
                    var _this = this;
                    if (!this.token) {
                        this.id = app.getLanguage().toUpperCase() + thatlist.util.id("xxxxxxxxxxxxxxxx", "ABCDEFGHJKLMNPQRSTUVWXYZ");
                        var connectPromise = thatlist.util.request(basePath + "/api/list/" + this.id + "/publish", "POST");
                        connectPromise.then(function (data) {
                            _this.token = data.token;
                        }, function (err) {
                            window.logError("error: can not publish list: ", err);
                        });
                        return connectPromise;
                    }
                    else {
                        return Promise.resolve();
                    }
                };
                List.prototype.invite = function () {
                    var b64Title = window.btoa(unescape(encodeURIComponent(this.title)));
                    var subject = $translate.instant("list_connect_msg_subject", { "listtitle": this.title });
                    var msg = $translate.instant("list_connect_msg", { "listtitle": this.title, "hash": window.btoa("_hex__") + window.btoa(this.id + "!" + this.token + "!" + b64Title) });
                    log(msg);
                    return thatlist.util.share(msg, subject, $translate.instant("connect_share_title"));
                };
                List.prototype.disconnect = function () {
                    this.token = void (0);
                    //gen add commands for current items
                    for (var i = 0; i < this.items.length; i++) {
                        var curItem = this.items[i];
                        this.addCommand({
                            "cmd": "add",
                            "good": curItem
                        });
                    }
                };
                List.prototype.sync = function () {
                    var _this = this;
                    return new Promise(function (resolve, reject) {
                        // if (this.tmp.syncing) return;
                        // this.tmp.syncing = true;
                        if (_this.needsSync())
                            _this.commit().then(resolve, reject);
                        else
                            _this.update().then(resolve, reject);
                        // setTimeout(() => {
                        //     this.tmp.syncing = null;
                        // }, throttle || 0);
                    });
                };
                List.prototype.needsSync = function () {
                    return this.token && this.commands.length > 0;
                };
                List.prototype.commit = function (isPauseing) {
                    var _this = this;
                    var commandsToTrans;
                    var syncPromise = new Promise(function (resolve, reject) {
                        //if shared TODO and online
                        if (_this.needsSync()) {
                            // this.tmp.commiting = this.tmp.updating = true;
                            var now = Date.now();
                            _this.commands.forEach(function (cmd, idx) { if (!cmd.id)
                                cmd.id = now + "_" + idx; });
                            // var waitForPromise;
                            if (isPauseing) {
                                //save with command id's
                                //waitForPromise = app.save(true);
                                app.save(true);
                            }
                            // else {
                            //     waitForPromise = Promise.resolve();
                            // }
                            // waitForPromise.then(() => {
                            commandsToTrans = JSON.parse(JSON.stringify(_this.commands));
                            //TODO what if something causes save during sync and app is killed before sync completes -> commands are lost
                            _this.commands = [];
                            // thatlist.util.request(
                            //     baseHost + "/!" + this.token + "/api/list/" + this.id + "/commit?version=" + appversion,
                            //     "POST",
                            //     {
                            //         "commands": commandsToTrans,
                            //         "oldVersion": this.remoteVersion
                            //     }
                            // );
                            //exec native to avoid killing by ios on pause
                            var xhr = new XMLHttpRequest();
                            xhr.open("POST", baseHost + "/!" + _this.token + "/api/list/" + _this.id + "/commit?version=" + window.appversion + (cordova.plugins && cordova.plugins.backgroundMode && cordova.plugins.backgroundMode.isActive() ? "&bg=true" : ""), true);
                            xhr.setRequestHeader("Content-Type", "application/json");
                            xhr.onload = function () {
                                log("commit result " + _this.ts + "  " + xhr.status);
                                if (xhr.status === 200) {
                                    var remotedata = JSON.parse(xhr.responseText);
                                    if (remotedata.version)
                                        _this.remoteVersion = remotedata.version;
                                    if (remotedata.items) {
                                        _this.setRemoteItems(remotedata.items);
                                        _this.sort();
                                    }
                                    if (remotedata.pickedItemsVersion)
                                        _this.remotePickedItemsVersion = remotedata.pickedItemsVersion;
                                    if (remotedata.pickedItems) {
                                        _this.setRemotePickedItems(remotedata.pickedItems);
                                    }
                                    // this.tmp.commiting = this.tmp.updating = false;
                                    app.save(true);
                                    resolve();
                                }
                                else {
                                    reject();
                                }
                            };
                            xhr.onerror = function () {
                                log("commit error");
                                reject();
                            };
                            xhr.ontimeout = function () {
                                log("commit timeout");
                                reject();
                            };
                            xhr.send(JSON.stringify({
                                "commands": commandsToTrans,
                                "oldVersion": _this.remoteVersion,
                                "oldPickedItemsVersion": _this.remotePickedItemsVersion
                            }));
                            // }, reject)
                        }
                        else {
                            resolve();
                        }
                    });
                    syncPromise.then(null, function () {
                        if (commandsToTrans && commandsToTrans.length > 0) {
                            //put commands back
                            _this.commands = _this.commands.concat(commandsToTrans);
                        }
                        // this.tmp.commiting = this.tmp.updating = false;
                        app.save();
                    });
                    return syncPromise;
                };
                List.prototype.update = function () {
                    var _this = this;
                    return new Promise(function (resolve, reject) {
                        //if share and shared (id) and network
                        if (_this.token) {
                            log(_this.title + " update " + _this.id + "  " + _this.remoteVersion);
                            if (_this.tmp.updating)
                                return;
                            if (_this.needsSync()) {
                                //exec failed commit instead
                                _this.commit().then(resolve, reject);
                                return;
                            }
                            _this.tmp.updating = true;
                            thatlist.util.request(baseHost + "/!" + _this.token + "/api/list/" + _this.id, "GET", { "oldVersion": _this.remoteVersion, "oldPickedItemsVersion": _this.remotePickedItemsVersion, "version": window.appversion }).then(function (remotedata, status) {
                                log(_this.title + " local version expired -> use received remotedata " + status + "  " + JSON.stringify(remotedata));
                                _this.tmp.updating = false;
                                if (remotedata.version)
                                    _this.remoteVersion = remotedata.version;
                                _this.setRemoteItems(remotedata.items);
                                //apply unsynced commands
                                _this.applyCommands();
                                if (remotedata.pickedItemsVersion)
                                    _this.remotePickedItemsVersion = remotedata.pickedItemsVersion;
                                _this.setRemotePickedItems(remotedata.pickedItems);
                                //sort and save
                                _this.sort();
                                resolve();
                            }, function (status) {
                                _this.tmp.updating = false;
                                if (status >= 400 || status === 0) {
                                    reject();
                                    log(_this.title + " can not read remotedata: " + status);
                                }
                                else {
                                    //e.g. not modified
                                    resolve();
                                    log(_this.title + " read remotedata: " + status);
                                }
                            });
                        }
                    });
                };
                /**
                 * used by sync only because commands are not updated
                 */
                List.prototype.setRemoteItems = function (items) {
                    var _this = this;
                    if (!items)
                        return;
                    log("setRemoteItems");
                    items.forEach(function (item) {
                        _this.checkCategory(item);
                        var curItemIdx = thatlist.util.indexOfItem(item.name, _this.items);
                        //Add missing and update (may be category was changed) existing favs. Keep prio when item in list already.
                        _this.addFavourite(item, { "keepPrio": curItemIdx >= 0 });
                    });
                    this.items = items;
                };
                List.prototype.setRemotePickedItems = function (remotePickedItems) {
                    if (!remotePickedItems)
                        return;
                    log("setRemotePickedItems");
                    var items = [];
                    for (var name in remotePickedItems) {
                        var item = remotePickedItems[name];
                        item.name = name;
                        this.checkCategory(item);
                        //add missing and update existing favs 
                        this.addFavourite(item, { "keepPrio": true });
                        items.push(item);
                    }
                    items.sort(function (a, b) { var ad = a.date || ""; var bd = b.date || ""; return ad < bd ? 1 : ad === bd ? 0 : -1; });
                    items.forEach(function (item) { delete item.date; });
                    this.pickedItems = items;
                };
                List.prototype.itemsToText = function (items) {
                    if (!items)
                        items = this.items;
                    items.sort(this.sortItemsByCategory);
                    this.sort();
                    var text = [];
                    var curCat;
                    for (var i = 0; i < items.length; i++) {
                        var curItem = items[i];
                        if (curCat != curItem.cat) {
                            curCat = curItem.cat;
                            text.push("\n[" + app.categories[curItem.cat].title + "]");
                        }
                        text.push((curItem.count ? curItem.count + " " : "") + curItem.name);
                    }
                    return text.join("\n");
                };
                return List;
            }());
            thatlist.List = List;
        })(thatlist = markenapps.thatlist || (markenapps.thatlist = {}));
    })(markenapps = com.markenapps || (com.markenapps = {}));
})(com || (com = {}));
