module com.markenapps.thatlist {
    export class App {
        version: string;

        lists: List[];
        /**
         * selected list local id
         */
        selectedListTS: number

        /**
         * custom categories (created or updated)
         */
        categories: {
            [name: string]: Category
        }

        date: string;
        runningtime: number = 0;
        backupDate: string;

        purchases: any[];

        settings: any;

        tmp: any = {
            catalog: [],
            units: []
        }


        constructor(data, assets) {
            if (data.version !== window.appversion) {
                log("update version from " + data.version);
                this.version = window.appversion;
            }

            this.date = data.date || new Date().toJSON();
            this.runningtime = data.runningtime || 0;
            this.backupDate = data.backupDate;

            //validate purchases
            let validatedPurchases = [];
            if (data.purchases) {
                data.purchases.forEach((purchase) => {
                    if (App.isProductValid(purchase)) {
                        validatedPurchases.push(purchase);
                    }
                });
            }
            this.purchases = validatedPurchases;

            let goods = assets.goods;
            for (let g = 0; g < goods.length; g++) {
                let curItem = goods[g];

                this.tmp.catalog.push({
                    "cat": curItem[0],
                    "name": curItem[1],
                    "prio": curItem[2] || 0
                });
            }


            if (data.categories) {
                for (var catName in data.categories) {
                    let curAssetCatKeywords = (assets.categories[catName] && assets.categories[catName].keywords) || [];

                    //override with custom categories
                    let curCat = assets.categories[catName] = data.categories[catName];

                    curCat.keywords = curAssetCatKeywords.map(keyword => thatlist.util.normalizeStr(keyword));
                }
            }
            this.categories = assets.categories;

            let units = assets.units;
            for (let u = 0; u < units.length; u++) {
                let curUnit = units[u];
                //TODO do not normalize here
                if (!curUnit.name) curUnit.name = curUnit.title;
                curUnit.normTitle = thatlist.util.normalizeStr(curUnit.title.split(" (")[0]);
                if (curUnit.title_sing) curUnit.normTitleSing = thatlist.util.normalizeStr(curUnit.title_sing.split(" (")[0]);
            }
            this.tmp.units = units;

            this.settings = data.settings || {};
            if (!this.settings.lang) this.setLanguage(thatlist.util.getFirstBrowserLanguage());
            if (!this.settings.fontSize) this.settings.fontSize = 3;
            if (this.settings.tips_hide === void (0)) this.settings.tips_hide = false;


            let local_theme_name = localStorage.getItem("theme-name");
            let local_theme_primaryColor = localStorage.getItem("theme-primaryColorIdx");
            let local_theme_accentColor = localStorage.getItem("theme-accentColorIdx");

            if (!this.settings.theme_name) this.settings.theme_name = local_theme_name;
            if (!this.settings.theme_primaryColor) this.settings.theme_primaryColor = local_theme_primaryColor;
            if (!this.settings.theme_accentColor) this.settings.theme_accentColor = local_theme_accentColor;

            if (local_theme_name !== this.settings.theme_name || local_theme_primaryColor !== !this.settings.theme_primaryColor || local_theme_accentColor !== !this.settings.theme_accentColor) {
                //theme correction
                setTheme(this.settings.theme_primaryColor, this.settings.theme_accentColor, this.settings.theme_name === "dark");
            }
            if (!data.lists) {
                //disable for new installation by default
                this.settings.lock_screen_feature = false;
            } else if (this.settings.lock_screen_feature === void (0)) {
                this.settings.lock_screen_feature = true;
            }

            if (this.settings.force_portait) thatlist.util.setOrientation('portrait');

            if (this.settings.extract_count === void (0)) this.settings.extract_count = true;

            this.lists = data.lists || [];
            if (this.lists.length === 0) {
                this.addList();
            }
            this.selectedListTS = data.selectedListTS || this.lists[0].ts;
        }

        static getSupportedLanguages(): any {
            return {
                'da': 'Dansk',
                'de': 'Deutsch',
                'en': 'English',
                'el': 'Ελληνικά',
                'es': 'español',
                'fr': 'français',
                'it': 'italiano',
                'nl': 'Nederlands',
                'no': 'norsk',
                'pl': 'polski',
                'pt': 'português',
                'fi': 'suomi',
                'sv': 'svenska',
                'cs': 'čeština',
                'ru': 'русский',
                'ja': '日本語'
            };
        }

        static isProductValid(product) {
            let valid = allProductIds.indexOf(product.productId || product.id) >= 0;

            if (!valid) log("product invalid " + JSON.stringify(product));

            return valid;
        }

        setLanguage(lang: string): any {
            this.settings.lang = thatlist.App.getSupportedLanguages()[lang] ? lang : '';
        }
        getLanguage(): string {
            return app.settings.lang || thatlist.util.getFirstBrowserLanguage()
        }

        /**
         * @returns true in case pruchases are loaded but there are no purchases
         */
        hasAds() {
            return this.purchases && !this.purchases.length;
        }

        onOwnedPurchase(handler) {
            if (!this.purchases) this.purchases = [];

            thatlist.util.onOwnedPurchase(
                allProductIds,
                (ownedProduct) => {
                    if (ownedProduct) {
                        log("ownedProduct " + JSON.stringify(ownedProduct));
                        if (!App.isProductValid(ownedProduct)) return;

                        this.purchases = [ownedProduct];

                        app.save();

                        handler(this.purchases);
                    }
                }
            );
        }

        buyProduct(productId: string): any {
            return new Promise((resolve, reject) => {
                thatlist.util.buyProduct(productId).then(
                    (purchase) => {
                        if (!purchase) {
                            reject();
                            return;
                        }

                        log("add purchase " + JSON.stringify(purchase));
                        this.purchases = [purchase];

                        this.save();

                        resolve(purchase);
                    },
                    () => {
                        reject();
                    }
                );
            });
        }

        addList() {
            let catOrder = Object.keys(this.categories);
            //filter custom categories title_iconname_color-hex-code
            catOrder = catOrder.filter(catId => { return catId.length <= 10 })
            let defOrder = ["FuV", "Ca", "SuD", "I", "D", "B", "S", "CuT", "C", "P", "Cu", "MuF", "Fish", "F", "R", "Sn", "DuH", "DIY", "O"];
            catOrder.sort((id1, id2) => {
                let prio1 = defOrder.indexOf(id1);
                if (prio1 < 0) prio1 = defOrder.length;
                let prio2 = defOrder.indexOf(id2);
                if (prio2 < 0) prio2 = defOrder.length;

                if (prio1 > prio2) return 1;
                if (prio1 === prio2) return 0;
                return -1;
            })

            let newList = new List({ "catOrder": catOrder });
            this.lists.push(newList);

            return newList;
        }

        cloneList(list) {
            let data = JSON.parse(JSON.stringify(list));
            let newList = this.addList();
            data.ts = Date.now();
            data.title = list.title + "2";

            for (let prop in data) {
                console.log(prop);
                newList[prop] = data[prop];
            }

            newList.disconnect();

            return newList;
        }

        getList(ts) {
            return this.lists.find(list => list.ts === ts);
        }

        selectList(ts) {
            log("Build new list!!!")

            var list = new List(this.lists.find(list => list.ts == ts) || this.lists[0]);
            this.selectedListTS = list.ts;

            return list;
        }

        deleteList(list) {
            let listIdx = this.indexOfList(list);
            if (listIdx >= 0) {
                this.lists.splice(listIdx, 1);
            }
        }

        indexOfList(list) {
            for (let l = 0; l < this.lists.length; l++) {
                if (this.lists[l].ts === list.ts) {
                    return l;
                }
            }

            return -1;
        }


        getFilteredUnits(parsedItem: any): any[] {
            if (!this.tmp.units || !this.tmp.units.length) return [];

            let filter = thatlist.util.normalizeStr(parsedItem.unit);
            if (filter == "x") filter = "";

            let matchedUnits = this.tmp.units.filter((unit) => {
                let unitName = unit.normTitle;
                let prio = 0;

                if (unitName == filter) {
                    prio = 1000;
                } else {
                    let idx = unitName.indexOf(filter);
                    if (idx >= 0) {
                        prio = idx == 0 ? 800 : idx > 0 ? 600 : 0;
                        prio -= unitName.length;
                    }
                }

                if (unit.dP) {
                    prio += unit.dP + unitName.length;
                }

                if (!unit.tmp) unit.tmp = {};
                unit.tmp.prio = prio;
                return prio;
            }).sort((a: any, b: any) => { return b.tmp.prio - a.tmp.prio });

            return matchedUnits;
        }

        hasPurchase(productId) {
            for (let p = 0; p < this.purchases.length; p++) {
                let curPurchase = this.purchases[p];
                if (curPurchase.productId === productId) {
                    return true;
                }
            }

            return false;
        }

        save(dataOrSave?) {
            let persistentData;

            if (dataOrSave && dataOrSave !== true) {
                log("save toJSON 1")
                persistentData = typeof (dataOrSave) === "string" ? dataOrSave : thatlist.util.toJSON(dataOrSave);
            } else {
                if (vm.list) {
                    let listIdx = this.indexOfList(vm.list);
                    if (listIdx >= 0) {
                        //replace ... otherwise primitive values will be lost because they are passed by copy and not be reference
                        //use splice ... otherwise vue will not recognize the change
                        this.lists.splice(listIdx, 1, vm.list);
                    }
                }
                log("save toJSON 2")
                persistentData = thatlist.util.toJSON(this);
            }

            if (writeThrottle) {
                clearTimeout(writeThrottle);
                writeThrottle = null;
            }

            //if we save other data -> take care the we are not killed by next save
            if (dataOrSave) {
                thatlist.util.writeFile("shoppinglist.json", persistentData, true);
                return Promise.resolve();
            } else {
                return new Promise((resolve, reject) => {
                    writeThrottle = setTimeout(() => {
                        thatlist.util.writeFile("shoppinglist.json", persistentData, true).then(resolve, reject);
                    }, 100);
                })
            }
        }


        hasCategory(catName) {
            return !catName || this.categories[catName];
        }

        /**
         * check if category needs to be created
         * @param catName 
         * 
         * @returns true if category was created otherwise false
         */
        checkCategoryName(catName) {
            if (!this.hasCategory(catName)) {
                //check if catName holds categorie properties
                var catNameParts = catName.split("_");

                if (catNameParts.length === 3) {
                    var catTitle = catNameParts[0];
                    var catIconName = "icon-" + catNameParts[1];
                    var catColor = "#" + catNameParts[2];

                    if (catColor.length === 7 && iconNames.indexOf(catIconName) >= 0) {
                        var newCat = {
                            "name": catName,
                            "title": catTitle,
                            "icon": catIconName,
                            "color": catColor,
                            "persist": true
                        };
                        this.categories[catName] = newCat;

                        return true;
                    }
                }
            }

            return false;
        }

        rate() {
            log("TODO check rate due");
        }

        doBackup(force?): Promise<any> {
            console.log(force, this.backupDate);
            if (force || !this.backupDate || Date.now() - new Date(this.backupDate).getTime() > 1000 * 60 * 60 * 12) {
                console.log("a");
                return new Promise((resolve, reject) => {
                    try {
                        var purchasesBck = this.purchases;
                        delete this.purchases;

                        var data: any = thatlist.util.toJSON(this);
                        data = data.replace(/\"commands\":/g, "\"_commands_\":");
                        data = JSON.parse(data);
                        data.backupDate = new Date().toJSON();
                        this.purchases = purchasesBck;

                        var zip = new JSZip();
                        //add year and month to delete outdated backups on backend
                        console.log(data);
                        zip.file("_tsl_" + (data.backupDate.substr(0, 7)) + ".json", thatlist.util.toJSON(data));
                        zip.generateAsync({
                            type: "blob",
                            compression: "DEFLATE",
                            compressionOptions: {
                                level: 9
                            }
                        }).then(
                            content => {
                                thatlist.util.requestBlob(getBackupPath(), "POST", content).then(
                                    () => {
                                        this.backupDate = data.backupDate;
                                        this.save();
                                        resolve();
                                    },
                                    reject
                                );
                            },
                            reject
                        );
                    } catch (e) {
                        log(e);
                        reject();
                    }
                });
            } else {
                return Promise.resolve()
            }
        }

        loadBackup(backupId, noWarn?) {
            return new Promise((resolve, reject) => {
                thatlist.util.requestBlob(getBackupPath(backupId))
                    .then(JSZip.loadAsync, (status) => {
                        if (status === 404) { this.backupDate = void (0); reject(); }
                    })
                    .then((zip: any) => {
                        return zip.file(Object.keys(zip.files)[0]).async("string");
                    }, reject)
                    .then(
                        text => {
                            try {
                                var data = JSON.parse(text);
                                resolve(data);
                            } catch (e) {
                                reject();
                            }
                        },
                        reject
                    );
            })
        }

        deleteBackup(backupId?) {
            return new Promise((resolve, reject) => {
                let bckDelPromise = thatlist.util.request(getBackupPath(backupId), "DELETE").then(
                    () => {
                        this.backupDate = void (0);
                        resolve();
                    },
                    reject
                );
            });
        }
    }

    export class List {
        /**
         * local id
         */
        ts: number;
        title: string;
        /**
         * items in list
         */
        items: Item[]
        pickedItems: Item[]
        pickedItemsLimit: number = 30;

        catOrder: string[]

        favourites: Favourite[]

        /**
         * remote id
         */
        id: string;
        token: string;
        remoteVersion: number;
        remotePickedItemsVersion: number;
        commands: Command[];

        tmp: any = {};

        sortItemsByCategory = (a, b) => {
            var pos = this.catOrder.indexOf(a.cat) - this.catOrder.indexOf(b.cat);

            if (pos != 0) return pos;

            return thatlist.util.normalizeStr(a.name) < thatlist.util.normalizeStr(b.name) ? -1 : 1;
        }

        constructor(data) {
            this.ts = data.ts || Date.now();
            this.title = data.title || $translate.instant("supermarket");

            //migrate do not persist
            //this.itemNames = data.itemNames || {};

            this.items = data.items || [];//(data.items || []).map(item => {item.tmp = {collapsed: false}; return item;});
            // this.items.forEach(item => {if (item && item.name) this.tmp.itemsByNames[item.name] = item});
            this.pickedItems = data.pickedItems || [];
            this.pickedItems = this.pickedItems.splice(0, 30);
            if (data.pickedItemsLimit !== void (0)) this.pickedItemsLimit = data.pickedItemsLimit;

            //clear inconsistent catOrder for dublicates and null
            var clearedCatOrder;
            if (data.catOrder) {
                clearedCatOrder = [];
                var cats = {};
                data.catOrder.forEach((cat) => {
                    if (!cat) return;

                    if (!cats[cat]) {
                        cats[cat] = true;
                        clearedCatOrder.push(cat);
                    }
                });
            }

            this.catOrder = clearedCatOrder;

            this.id = data.id || "";
            this.token = data.token || "";
            this.remoteVersion = data.remoteVersion;
            this.commands = data.commands || [];

            let favsByName = {};
            if (data.favourites) {
                if (!Array.isArray(data.favourites)) {
                    let favourites: Favourite[] = [];
                    //migrate to array
                    for (let name in data.favourites) {
                        let curFav: Favourite = data.favourites[name];
                        if (!curFav || !name) continue;

                        //migrate use [data, interval, use] -> prio
                        favourites.push({
                            "name": name,
                            "cat": curFav.cat,
                            "prio": curFav.prio || curFav.use[2] || 1
                        });
                    }
                    data.favourites = favourites;
                }

                let doDowngrade = false;
                let recentDowngradeTS = parseInt(localStorage.getItem("recentDowngradeTS") || "0", 10);
                //downgrade prio once a day only
                if (!recentDowngradeTS || Date.now() - recentDowngradeTS > 60000 * 60 * 24) {
                    localStorage.setItem("recentDowngradeTS", Date.now());
                    doDowngrade = true;
                }

                data.favourites.forEach(fav => {
                    //fix invisible favs
                    if (!fav.prio) fav.prio = 1;

                    favsByName[fav.name] = fav;
                    //naive downgrade prio/use count 
                    if (doDowngrade && fav.prio > 4) {
                        //fav with prio 4 is downgraded by 1 when it's not added on 50 usage days
                        fav.prio = Math.max(4, fav.prio - (fav.prio / 200));
                        //round for two decimals
                        fav.prio = Math.round(fav.prio * 100) / 100;
                    }
                })
            }
            this.favourites = data.favourites || [];

            //because we init fist list in constuctor of app
            if (app && app.tmp && app.tmp.catalog) {
                let catalog = JSON.parse(JSON.stringify(app.tmp.catalog)).filter(item => { return !favsByName[item.name] });
                this.tmp.catalog = catalog.concat(this.favourites);
            }
        }

        private addCommand(command: Command) {
            if (command.cmd && command.picked) throw "command can have either a cmd or a picked property";

            if (command && command.cmd && command.good) {
                //handle conflicts
                for (let c = this.commands.length; c--;) {
                    const curCmd = this.commands[c];

                    if (command.good.name === curCmd.good.name) {
                        //dublicate
                        if (curCmd.cmd === command.cmd) {
                            //update old with new 
                            this.commands[c] = command;
                            return;
                        } else if (curCmd.cmd === "add" && command.cmd === "del") {
                            // TODO ignore both when add is unsynced
                            this.commands.splice(c, 1);
                            break;
                        } else if (curCmd.cmd === "del" && command.cmd === "add") {
                            //ignore del and apply add (may be count differs)
                            this.commands.splice(c, 1);
                            break;
                        }
                    }
                }

                this.commands.push({
                    "cmd": command.cmd,
                    "good": {
                        "cat": command.good.cat,
                        "name": command.good.name,
                        "count": command.good.count
                    }
                });
            }

            if (command && command.picked && command.good) {
                //handle conflicts
                for (let c = this.commands.length; c--;) {
                    const curCmd = this.commands[c];

                    //backward compatibility
                    if (curCmd.cmd) continue;

                    if (command.good.name === curCmd.good.name) {
                        //dublicate
                        if (curCmd.picked === command.picked) {
                            //update old with new 
                            this.commands[c] = command;
                            return;
                        } else if (curCmd.picked === "add" && command.picked === "del") {
                            //ignore add and apply del TODO ignore both when add is unsynced
                            this.commands.splice(c, 1);
                            break;
                        } else if (curCmd.picked === "del" && command.picked === "add") {
                            //ignore del and apply add (may be count differs)
                            this.commands.splice(c, 1);
                            break;
                        }
                    }
                }

                this.commands.push({
                    "picked": command.picked,
                    "name": command.name,
                    "good": {
                        "cat": command.good.cat,
                        "name": command.good.name,
                        "count": command.good.count
                    }
                });
            }

            log(thatlist.util.toJSON(this.commands));
        }

        private applyCommands(options?) {
            this.commands.forEach(command => {
                var item = command.good;
                if (!item || !item.name) return;

                if (command.cmd === "add") {
                    this.checkCategory(item)

                    log("add ", item)

                    this.addItem(item, options);
                } else if (command.cmd === "del") {
                    let curItemIdx = thatlist.util.indexOfItem(item.name, this.items);

                    if (curItemIdx >= 0) {
                        this.items.splice(curItemIdx, 1);
                    }
                } else if (command.cmd === "update") {
                    //if not there do not insert ... may be it was deleted
                    var idx = thatlist.util.indexOfItem(item.name, this.items);
                    if (idx >= 0) {
                        this.items.splice(idx, 1, item);
                    } else {
                        return;
                    }
                }
            })
        }

        private getFavouriteByName(itemName): Favourite {
            return this.favourites[thatlist.util.indexOfItem(itemName, this.favourites)];
        }

        private getItemByName(itemName) {
            return this.items[thatlist.util.indexOfItem(itemName, this.items)];
        }

        private hasCategory(catName) {
            return this.catOrder.indexOf(catName) >= 0;
        }

        hideCategory(category: Category) {
            let idx = this.catOrder.indexOf(category.name);

            if (idx >= 0) this.catOrder.splice(idx, 1);
        }

        unhideCategory(category: Category) {
            let idx = this.catOrder.indexOf(category.name);

            if (idx < 0) this.catOrder.push(category.name);
        }

        /**
         * Detect category if non is present and add category if missing or hidden
         * @param item 
         */
        public checkCategory(item: Item) {
            let origCat = item.cat;

            //detailed check category (based on catalog)
            if (!item.cat) {
                let g;
                try {
                    let bestCatRank = 100;

                    for (g = 0; g < this.tmp.catalog.length; g++) {
                        let curItem = this.tmp.catalog[g];

                        //check if cur good categorie is a visible categorie
                        if (!this.hasCategory(curItem.cat)) continue;

                        ////remove plural
                        let itemSearchName = thatlist.util.normalizeStr(item.name);//.length>=4 ? good.normName.substring(0, good.normName.length-2) : good.normName;
                        let curItemSearchName = thatlist.util.normalizeStr(curItem.name);//.length>=4 ? curGood.normName.substring(0, curGood.normName.length-2) : curGood.normName;
                        if (!curItemSearchName) {
                            continue;
                        }

                        let curCatRank = Math.abs(curItemSearchName.length - itemSearchName.length);

                        if (curCatRank > bestCatRank) continue;

                        if (curItem && curItemSearchName && curItem.cat && curItem.cat !== "O") {
                            //avoid öl (ol) match rucola but let ei match eier
                            if (curItemSearchName.indexOf(itemSearchName) >= 0) {
                                log(".cat= 3");
                                item.cat = curItem.cat;
                                bestCatRank = curCatRank;

                                if (curCatRank === 0) break;
                            } else if (itemSearchName.indexOf(curItemSearchName) >= 0) {
                                log(".cat= 4");
                                item.cat = curItem.cat;
                                bestCatRank = curCatRank;

                                if (curCatRank === 0) break;
                            }
                        }
                    }
                } catch (e) {
                    window.logError("APP ERROR get cat: " + e.message + (e.stack ? "\n" + e.stack : "") + "-" + this.tmp.catalog + "-" + item.name + "-" + this.tmp.catalog[g].name);
                }
            }

            //fallback detect category (based on keywords)
            if (!item.cat) {// || item.cat === "O") {
                delete item.cat;

                //detect cat
                let bestCat;
                let bestCatPrio = -1
                this.catOrder.forEach(catName => {
                    let curCatPrio = 0;

                    let filter = thatlist.util.normalizeStr(item.name);

                    let curCat = app.categories[catName];
                    if (curCat && filter) {
                        for (let k = 0; curCat.keywords && k < curCat.keywords.length; k++) {
                            let curKeyword = curCat.keywords[k];

                            if (filter.indexOf(curKeyword) >= 0) {
                                curCatPrio = curKeyword.length;
                                break;
                            }
                        }
                    }

                    if (curCatPrio && curCatPrio >= bestCatPrio) {
                        bestCat = curCat;
                        bestCatPrio = curCatPrio;
                    }
                });

                if (bestCat) {
                    log(".cat= 2");
                    item.cat = bestCat.name;
                }
            }

            if (!item.cat) {
                log(".cat= 5");
                item.cat = origCat || "O";
            }

            app.checkCategoryName(item.cat);

            if (item.cat && !this.hasCategory(item.cat)) this.catOrder.push(item.cat);
        }

        addItem(item: Item, options?) {
            if (!item || !item.name) {
                return;
            }

            if (!options) options = {};

            if (!item.name) return;
            item.name = item.name.trim();

            this.checkCategory(item);

            if (item.tmp && item.tmp.count) {
                item.count = item.tmp.count;
                delete item.tmp.count;
            }

            // check if this item is already in list
            let curItemIdx = thatlist.util.indexOfItem(item.name, this.items);

            // add/update command
            this.addCommand({
                "cmd": "add",
                "good": item,
            })

            if (curItemIdx >= 0) {
                //do update prio when not already in items list
                if (typeof options.keepPrio === void (0)) options.keepPrio = true;

                //replace item
                this.items.splice(curItemIdx, 1, item)[0];
            } else {
                if (!options.bottom) this.items.unshift(item);
                else this.items.push(item);
            }

            //add/update fav (may be cat or count was changed)
            this.addFavourite(item, options);

            if (options.sort || options.sort === void (0)) this.sort();
        }

        addItems(items, options: any = {}) {
            items.forEach((item) => {
                //add item only if not in list already
                if (!this.getItemByName(item.name)) {
                    this.addItem(item, { "sort": false });
                }
            });

            if (options.clearPicked) {
                this.pickedItems = [];
            }

            this.sort();
        }

        parseItemsFromSpeech(text) {
            var units = app.tmp.units;
            var result = text;
            var items = result.split(thatlist.util.andSplitWords[app.getLanguage()]);
            var itemsCollection = [];
            items.forEach((item) => {
                var cleanItemText = item.trim();
                if (!cleanItemText) return;
                var itemParts = cleanItemText.split(" ");
                var precWasNumber;
                for (var p = 0; p < itemParts.length; p++) {
                    var part = itemParts[p];
                    var normPart = part.toLowerCase();
                    var isUnit = false;
                    var num = isNaN(part) ? thatlist.util.numberWordToNumber(normPart) : part;
                    if (num) {
                        itemParts[p] = num;

                        precWasNumber = true;
                    } else if (precWasNumber) {
                        precWasNumber = false;

                        //check if unit
                        for (var u = 0; u < units.length; u++) {
                            var curUnit = units[u];
                            //curUnit.normTitle = curUnit.normTitle === void (0) ? curUnit.title.split(" (")[0].toLowerCase() : curUnit.normTitle;
                            //curUnit.normTitleSing = curUnit.title_sing && curUnit.normTitleSing === void (0) ? curUnit.title_sing.split(" (")[0].toLowerCase() : curUnit.normTitleSing;

                            if (normPart === curUnit.normTitle) isUnit = true;
                            else if (normPart === curUnit.normTitleSing) isUnit = true;

                            if (isUnit) {
                                //replace long unit with short unit name
                                if (curUnit.name) itemParts[p] = part = curUnit.name;
                                break;
                            }
                        }

                        //heuristik for unknown units
                        if (!isUnit && itemParts.length > 2 && p === itemParts.length - 1) {
                            //guess in case of "name count unit" e.g. Milch mit wenig Fett 3 Liter
                            isUnit = true;
                        }

                        //join amount and unit
                        if (isUnit) {
                            //push amount and unit to the start
                            itemParts.unshift(itemParts[p - 1] + part)
                            //remove from the end
                            itemParts.splice(p, 2);
                        }
                    }
                }

                var item = thatlist.util.parseItemName(itemParts.join(" "));
                if (item) {
                    // if (item.name && item.name.length > 1) item.selected = true;
                    itemsCollection.push(item);
                }
            })

            return itemsCollection;
        }


        removeItem(item) {
            let idx = thatlist.util.indexOfItem(item.name, this.items);
            if (idx >= 0) {
                this.addCommand({
                    "cmd": "del",
                    "good": item
                });

                return this.items.splice(idx, 1)[0];
            } else {
                return null
            }
        }

        clearItems() {
            if (this.token) {
                this.items.forEach(item => {
                    this.addCommand({
                        "cmd": "del",
                        "good": item
                    })
                });
            } else {
                //clear commands if list is not synced
                this.commands.length = 0;
            }

            this.items = [];
        }

        addPickedItem(item, options: any = {}) {
            if (!item || !item.name) return;

            if (!app.settings.keep_count) {
                if (!item.tmp) item.tmp = {};
                item.tmp.pickedCount = item.count;

                let curFav: Favourite = this.getFavouriteByName(item.name);
                if (curFav) delete curFav.count;

                delete item.count;
            }

            let insertAt = 0;
            if (options.replaceItem) {
                let idx = thatlist.util.indexOfItem(options.replaceItem.name, this.pickedItems);
                if (idx >= 0) {
                    insertAt = idx;
                    this.addCommand({ "picked": "replace", "name": options.replaceItem.name, "good": item });
                    this.pickedItems.splice(idx, 1);
                }
            } else {
                let idx = thatlist.util.indexOfItem(item.name, this.pickedItems);
                if (idx >= 0) this.pickedItems.splice(idx, 1);

                this.addCommand({ "picked": "add", "good": item });
                this.pickedItemsLimit++;
            }

            this.pickedItems.splice(insertAt, 0, item);
        }

        removePickedItem(item) {
            let curPickedItemIdx = thatlist.util.indexOfItem(item.name, this.pickedItems);

            if (curPickedItemIdx >= 0) {
                if (this.pickedItemsLimit && curPickedItemIdx < this.pickedItemsLimit) this.pickedItemsLimit--;

                this.addCommand({
                    "picked": "del",
                    "good": item
                });

                return this.pickedItems.splice(curPickedItemIdx, 1)[0];
            }
        }

        hidePickedItems() {
            this.pickedItemsLimit = 0;
        }


        /**
         * add/update fav if not already present
         * @param item item to add
         */
        addFavourite(item, options?: any) {
            if (!item || !item.name) return;

            if (!options) options = {};

            try {
                //check fav
                let curFav: Favourite = this.getFavouriteByName(item.name);

                if (!curFav) {
                    //add favourite
                    curFav = JSON.parse(JSON.stringify(item));
                    this.favourites.unshift(curFav);

                    //update catalog
                    let curItemIdx = thatlist.util.indexOfItem(item.name, this.tmp.catalog);
                    if (curItemIdx < 0) {
                        this.tmp.catalog.unshift(curFav);
                    } else {
                        //reuse cur prio of good
                        if (!curFav.prio) curFav.prio = this.tmp.catalog[curItemIdx].prio;
                        this.tmp.catalog.splice(curItemIdx, 1, curFav);
                    }
                } else {
                    log(".cat= 6", !!curFav, !!item);
                    curFav.cat = item.cat;
                    curFav.count = item.count;
                }

                if (!app.settings.keep_count) {
                    delete curFav.count;
                }

                let curUsedDate = new Date().toJSON().substring(0, 10);
                if (!curFav.tmp) curFav.tmp = {};
                let recentUsedDate = curFav.tmp.usedDate;

                if (!curFav.prio) {
                    curFav.prio = 3;
                } else {
                    //avoid double use by corretion: done->add from history->done again
                    if (!options.keepPrio && recentUsedDate !== curUsedDate) {
                        curFav.tmp.usedDate = curUsedDate;
                        if (curFav.prio < 10) curFav.prio++;
                    }
                }
            } catch (e) {
                window.logError("APP ERROR update fav: " + e.message + (e.stack ? "\n" + e.stack : ""))
            }
        }

        /**
         * remove item if present
         * @param item item to remove
         */
        removeFavourite(item) {
            if (!item || !item.name) return;
            let idx = thatlist.util.indexOfItem(item.name, this.favourites);
            if (idx >= 0) {
                //hide from search result until restart
                item.prio = 0;
                this.favourites.splice(idx, 1);
            }
        }

        deleteItem(item) {
            if (!item || !item.name) return;
            let idx = thatlist.util.indexOfItem(item.name, this.tmp.catalog);
            if (idx >= 0) {
                this.tmp.catalog.splice(idx, 1);
            }

            this.removeItem(item);
            this.removePickedItem(item);
            this.removeFavourite(item);
        }

        sort() {
            this.items.sort(this.sortItemsByCategory);

            app.save();
        }

        getCatalogItems(item) {
            let count = item.count;
            let filter = thatlist.util.normalizeStr(item.name);

            if (!filter) return [];

            let catsByName = {};
            this.catOrder.forEach(catName => { catsByName[catName] = true });

            let items = this.tmp.catalog.filter(item => {
                if (!catsByName[item.cat] || item.prio === 0) return false;

                //var goodCat = good.normCat || normalizeStr(good.cat);
                let itemName = thatlist.util.normalizeStr(item.name);
                let prio = 0;
                if (itemName == filter) {
                    prio = 1000;
                } else {
                    let idx = itemName.indexOf(filter);
                    if (idx >= 0) {
                        prio = idx == 0 ? 800 : idx > 0 ? 600 : 0;
                        if (idx > 0) {
                            prio -= itemName.length;
                        }
                    }
                    if (item.name === "Mouka") console.log("Mouka1 ", prio)
                    if (item.name === "Mléko") console.log("Mléko1 ", prio)
                }

                if (prio > 0 && item.prio) {
                    prio += item.prio;
                }

                if (item.name === "Mouka") console.log("Mouka2 ", prio)
                if (item.name === "Mléko") console.log("Mléko2 ", prio)

                if (!item.tmp) item.tmp = {}
                item.tmp.count = count;
                item.tmp.prio = prio;
                return prio;
            }).sort((a: any, b: any) => { return b.tmp.prio - a.tmp.prio }).slice(0, 8);

            return items;
        }


        getFavourites(orderBy, limit?) {
            if (limit === 0) return [];

            log("calc favorites by " + orderBy + " " + limit)
            let favourites = this.favourites.filter(item => {
                return thatlist.util.indexOfItem(item.name, this.items) < 0;
            });

            favourites.sort((a: Item, b: Item) => { return b.prio - a.prio });

            if (orderBy === "cat") {
                favourites.sort((a: Item, b: Item) => {
                    let asv = "0000" + this.catOrder.indexOf(a.cat);
                    asv = asv.substring(asv.length - 4) + thatlist.util.normalizeStr(a.name);
                    let bsv = "0000" + this.catOrder.indexOf(b.cat);
                    bsv = bsv.substring(bsv.length - 4) + thatlist.util.normalizeStr(b.name);

                    if (asv > bsv) return 1;
                    if (asv < bsv) return -1;
                    return 0;
                });
            }

            if (limit) favourites = favourites.splice(0, limit);

            return favourites;
        }

        connect(): Promise<any> {
            if (!this.token) {
                this.id = app.getLanguage().toUpperCase() + thatlist.util.id("xxxxxxxxxxxxxxxx", "ABCDEFGHJKLMNPQRSTUVWXYZ");

                let connectPromise = thatlist.util.request(
                    basePath + "/api/list/" + this.id + "/publish",
                    "POST"
                )

                connectPromise.then(
                    (data: any) => {
                        this.token = data.token;
                    },
                    err => {
                        window.logError("error: can not publish list: ", err);
                    }
                );

                return connectPromise;
            } else {
                return Promise.resolve();
            }
        }

        invite() {
            var b64Title = window.btoa(unescape(encodeURIComponent(this.title)));
            let subject = $translate.instant("list_connect_msg_subject", { "listtitle": this.title });
            let msg = $translate.instant("list_connect_msg", { "listtitle": this.title, "hash": window.btoa("_hex__") + window.btoa(`${this.id}!${this.token}!${b64Title}`) });

            log(msg);

            return thatlist.util.share(msg, subject, $translate.instant("connect_share_title"));
        }

        disconnect() {
            this.token = void (0);

            //gen add commands for current items
            for (var i = 0; i < this.items.length; i++) {
                var curItem = this.items[i];
                this.addCommand({
                    "cmd": "add",
                    "good": curItem
                })
            }
        }

        sync() {
            return new Promise((resolve, reject) => {
                // if (this.tmp.syncing) return;
                // this.tmp.syncing = true;

                if (this.needsSync()) this.commit().then(resolve, reject);
                else this.update().then(resolve, reject);

                // setTimeout(() => {
                //     this.tmp.syncing = null;
                // }, throttle || 0);
            })
        }

        needsSync() {
            return this.token && this.commands.length > 0;
        }

        commit(isPauseing?) {
            var commandsToTrans;
            let syncPromise = new Promise((resolve, reject) => {
                //if shared TODO and online
                if (this.needsSync()) {
                    // this.tmp.commiting = this.tmp.updating = true;

                    var now = Date.now();
                    this.commands.forEach((cmd, idx) => { if (!cmd.id) cmd.id = now + "_" + idx; })

                    // var waitForPromise;
                    if (isPauseing) {
                        //save with command id's
                        //waitForPromise = app.save(true);
                        app.save(true);
                    }
                    // else {
                    //     waitForPromise = Promise.resolve();
                    // }

                    // waitForPromise.then(() => {
                    commandsToTrans = JSON.parse(JSON.stringify(this.commands));
                    //TODO what if something causes save during sync and app is killed before sync completes -> commands are lost
                    this.commands = [];

                    // thatlist.util.request(
                    //     baseHost + "/!" + this.token + "/api/list/" + this.id + "/commit?version=" + appversion,
                    //     "POST",
                    //     {
                    //         "commands": commandsToTrans,
                    //         "oldVersion": this.remoteVersion
                    //     }
                    // );

                    //exec native to avoid killing by ios on pause
                    var xhr = new XMLHttpRequest();
                    xhr.open("POST", baseHost + "/!" + this.token + "/api/list/" + this.id + "/commit?version=" + window.appversion + (cordova.plugins && cordova.plugins.backgroundMode && cordova.plugins.backgroundMode.isActive() ? "&bg=true" : ""), true);
                    xhr.setRequestHeader("Content-Type", "application/json");
                    xhr.onload = () => {
                        log("commit result " + this.ts + "  " + xhr.status)
                        if (xhr.status === 200) {
                            var remotedata = JSON.parse(xhr.responseText);

                            if (remotedata.version) this.remoteVersion = remotedata.version;
                            if (remotedata.items) {
                                this.setRemoteItems(remotedata.items);
                                this.sort();
                            }

                            if (remotedata.pickedItemsVersion) this.remotePickedItemsVersion = remotedata.pickedItemsVersion;
                            if (remotedata.pickedItems) {
                                this.setRemotePickedItems(remotedata.pickedItems);
                            }

                            // this.tmp.commiting = this.tmp.updating = false;

                            app.save(true);
                            resolve();
                        } else {
                            reject();
                        }
                    };
                    xhr.onerror = () => {
                        log("commit error");
                        reject();
                    }
                    xhr.ontimeout = () => {
                        log("commit timeout");
                        reject();
                    }
                    xhr.send(JSON.stringify({
                        "commands": commandsToTrans,
                        "oldVersion": this.remoteVersion,
                        "oldPickedItemsVersion": this.remotePickedItemsVersion
                    }));
                    // }, reject)
                } else {
                    resolve();
                }
            });

            syncPromise.then(null, () => {
                if (commandsToTrans && commandsToTrans.length > 0) {
                    //put commands back
                    this.commands = this.commands.concat(commandsToTrans);
                }

                // this.tmp.commiting = this.tmp.updating = false;

                app.save();
            });

            return syncPromise;
        }

        update() {
            return new Promise((resolve, reject) => {
                //if share and shared (id) and network
                if (this.token) {
                    log(this.title + " update " + this.id + "  " + this.remoteVersion);

                    if (this.tmp.updating) return;

                    if (this.needsSync()) {
                        //exec failed commit instead
                        this.commit().then(resolve, reject);
                        return;
                    }

                    this.tmp.updating = true;
                    thatlist.util.request(
                        baseHost + "/!" + this.token + "/api/list/" + this.id,
                        "GET",
                        { "oldVersion": this.remoteVersion, "oldPickedItemsVersion": this.remotePickedItemsVersion, "version": window.appversion }
                    ).then(
                        (remotedata, status) => {
                            log(this.title + " local version expired -> use received remotedata " + status + "  " + JSON.stringify(remotedata));
                            this.tmp.updating = false;

                            if (remotedata.version) this.remoteVersion = remotedata.version;
                            this.setRemoteItems(remotedata.items);

                            //apply unsynced commands
                            this.applyCommands();

                            if (remotedata.pickedItemsVersion) this.remotePickedItemsVersion = remotedata.pickedItemsVersion;
                            this.setRemotePickedItems(remotedata.pickedItems);

                            //sort and save
                            this.sort();

                            resolve();
                        },
                        (status) => {
                            this.tmp.updating = false;

                            if (status >= 400 || status === 0) {
                                reject();
                                log(this.title + " can not read remotedata: " + status);
                            } else {
                                //e.g. not modified
                                resolve();
                                log(this.title + " read remotedata: " + status);
                            }
                        }
                    )
                }
            });
        }

        /**
         * used by sync only because commands are not updated
         */
        private setRemoteItems(items) {
            if (!items) return;

            log("setRemoteItems")
            items.forEach(item => {
                this.checkCategory(item);

                let curItemIdx = thatlist.util.indexOfItem(item.name, this.items);
                //Add missing and update (may be category was changed) existing favs. Keep prio when item in list already.
                this.addFavourite(item, { "keepPrio": curItemIdx >= 0 });
            })

            this.items = items;
        }

        private setRemotePickedItems(remotePickedItems) {
            if (!remotePickedItems) return;

            log("setRemotePickedItems")
            var items = [];
            for (var name in remotePickedItems) {
                var item = remotePickedItems[name];
                item.name = name;
                this.checkCategory(item);

                //add missing and update existing favs 
                this.addFavourite(item, { "keepPrio": true });

                items.push(item);
            }

            items.sort(function (a, b) { var ad = a.date || ""; var bd = b.date || ""; return ad < bd ? 1 : ad === bd ? 0 : -1; });

            items.forEach((item) => { delete item.date });

            this.pickedItems = items;
        }

        itemsToText(items?) {
            if (!items) items = this.items;

            items.sort(this.sortItemsByCategory);

            this.sort();
            var text = [];
            var curCat;
            for (var i = 0; i < items.length; i++) {
                var curItem = items[i];
                if (curCat != curItem.cat) {
                    curCat = curItem.cat;
                    text.push("\n[" + app.categories[curItem.cat].title + "]");
                }

                text.push((curItem.count ? curItem.count + " " : "") + curItem.name);

            }
            return text.join("\n");
        }
    }

    export interface Category {
        name: string
        title: string
        color: string
        //for color selection only
        colorName?: string
        icon: string
        keywords?: string[]
        pos?: number
        prio?: number
        persist: boolean
    }

    export interface Command {
        //cmd to apply to items
        cmd?: string
        //cmd to apply to picked items e.g. add
        picked?: string
        //name of item to replace
        name?: string
        good: Item
        /** for idempotence if in case the app was killed after commit and before save*/
        id?: string
    }

    export interface Item {
        cat: string
        name: string
        count?: string
        tmp?: any
        prio?: number
    }
    export interface Favourite extends Item {
        //deprecated ... migrated to prio
        use?: number
    }
}