module com.markenapps.thatlist.util {
    const replWSRegExp = /\s/g;
    const replBSRegExp = / +(?= )/g;
    const replBracketsRegExp = /\((.{1,2})\)/g;
    const replExplRegExp = /\((.{3,})\)/g;
    const replSlashNRegExp = /\/(n\s)/g;
    const replSlashSRegExp = /\/(s\s)/g;

    let writeThrottles = {};
    let devicePromise;
    let providerCheckPromise;

    let noop = function () { };

    export function promiseDeviceReady() {
        if (!devicePromise) {
            devicePromise = new Promise(function (resolve, reject) {
                document.addEventListener("deviceready", resolve, true);
            });
        }

        return devicePromise;
    }

    export function exit() {
        navigator.app.exitApp();
    }

    export function showKeyboard() {
        if (window.Keyboard && window.Keyboard.show) window.Keyboard.show();
        else log("Keyboard not available")
    }

    export function hideKeyboard() {
        if (window.Keyboard && window.Keyboard.hide) window.Keyboard.hide();
        else log("Keyboard not available")
    }

    let awakeTMO;
    var wakeLock;
    export function keepAwake(tmo = 600000) {
        if (!window.plugins || !window.plugins.insomnia) return;

        try {
            window.plugins.insomnia.keepAwake();
            if (window.navigator.requestWakeLock) {
                wakeLock = window.navigator.requestWakeLock('screen');
            }
            if (awakeTMO) clearTimeout(awakeTMO);
            awakeTMO = setTimeout(function () { allowSleep() }, tmo);
        } catch (e) {
            log("keepAwake failed ", e);
        }

    }
    export function allowSleep() {
        if (!window.plugins || !window.plugins.insomnia) return;
        try {
            window.plugins.insomnia.allowSleepAgain();
            if (wakeLock) wakeLock.unlock();
        }
        catch (e) {
            log("allowSleep failed ", e);
        }
    }

    export function setStatusbarHidden(hidden) {
        if (!window.StatusBar || !window.StatusBar.hide) return;
        try {
            if (hidden) {
                vm.$f7.statusbar.hide();
                window.StatusBar.hide();
            } else {
                vm.$f7.statusbar.show();
                window.StatusBar.show();
            }
        }
        catch (e) {
            log("hideStatusbar failed ", e);
        }
    }

    export function setOrientation(orientation?) {
        if (!screen || !screen.orientation) return;
        try {
            if (orientation) screen.orientation.lock(orientation);
            else screen.orientation.unlock();
        }
        catch (e) {
            log("lockOrientation failed ", e);
        }
    }

    export module clipboard {
        export function copy(text) {
            return new Promise((resolve, reject) => {
                if (cordova && cordova.plugins && cordova.plugins.clipboard) {
                    //always add to storage to allow paste fallback
                    copyFallback(text, noop, noop);
                    cordova.plugins.clipboard.copy(text, resolve, () => { copyFallback(text, resolve, reject) });
                } else {
                    log("clipboad not available")
                    copyFallback(text, resolve, reject);
                }
            });
        }

        export function paste() {
            return new Promise((resolve, reject) => {
                if (cordova && cordova.plugins && cordova.plugins.clipboard) {
                    /*
                        XXX requires fix in plugin /shoppinglist/plugins/com.verso.cordova.clipboard/src/android/Clipboard.java
                        uncoment mimetype check
                        //if (!clipboard.getPrimaryClipDescription().hasMimeType(ClipDescription.MIMETYPE_TEXT_PLAIN)) {
                        //    callbackContext.sendPluginResult(new PluginResult(PluginResult.Status.NO_RESULT));
                        //}
                    */
                    cordova.plugins.clipboard.paste(
                        (text) => { log(text); resolve(text) },
                        () => { pasteFallback(resolve, reject) }
                    );
                } else {
                    log("clipboad not available")
                    pasteFallback(resolve, reject);
                }
            });
        }

        function copyFallback(text, resolve, reject) {
            try {
                sessionStorage.setItem('clipboard', text);
                resolve();
            } catch (e) {
                log("copy fallback failed ", e);
                reject();
            }
        }

        function pasteFallback(resolve, reject) {
            try {
                resolve(sessionStorage.getItem('clipboard'))
            } catch (e) {
                log("paste fallback failed ", e);
                reject();
            }
        }
    }


    export function initStore(productIds): Promise<any> {
        return new Promise((resolve, reject) => {
            if (!window.store) reject();

            store.error(function (e) {
                log("store error " + e.code + ": " + e.message);
                reject();
            });

            store.verbosity = store.DEBUG;

            productIds.forEach((id) => {
                log('registerProduct' + id);

                store.register({
                    "id": id,
                    "alias": id,
                    "type": store.NON_CONSUMABLE
                });
            });

            log("store initialized")
            resolve();
        });
    }

    // function isPurchaseValid(productIds, purchase) {
    //     return purchase && (purchase.state === void (0) || purchase.state == 0) && productIds.indexOf(purchase.productId) >= 0;
    // }

    /**
     * 
     * @param productIds 
     * 
     * @returns A promise that resolves to an array of promises for each product or rejects when store fails to initialize. 
     *          The product promises resolves when a product is owned.
     */
    export function onOwnedPurchase(productIds, handler) {
        log("onOwnedPurchase");


        initStore(productIds).then(() => {
            log("store refresh")
            store.refresh();

            productIds.forEach((id) => {
                store.when(id).updated((product) => {
                    log("product updated " + id + " " + (product ? product.state : "-"));
                })
            });

            store.when("non consumable").owned((p) => {
                log("owned " + JSON.stringify(p))
                handler(p);
            });
            store.when("non consumable").approved((p) => {
                log("approved > finish " + p.id)
                p.finish();
            });
        })
    }


    /**
     * {
        "transactionId": "f1",
        "receipt": ""
      }
     */
    export function buyProduct(productId) {
        log("buyProduct");
        return new Promise((resolve, reject) => {
            initStore([productId]).then(
                () => {
                    log("order " + productId);

                    store.when(productId).updated((product) => {
                        log("product updated " + productId + " " + (product ? product.state : "-"));
                    });

                    store.refresh();

                    store.order(productId)
                        .then(
                            (product) => {
                                store.once(productId, "approved", (product) => {
                                    log("finish " + productId);
                                    product.finish();
                                    resolve(product);
                                });

                                store.once(productId, "error", (e: any) => {
                                    if (!e) e = {};
                                    logError("ERROR productPromise" + e.code + ": " + e.message);

                                    reject();
                                });
                            })
                        .error(
                            (e) => {
                                logError("order " + productId + " failed " + e.code + ": " + e.message);

                                reject();
                            }
                        )
                },
                (e) => {
                    reject();
                }
            );
        });

        // if (!inAppPurchase) {
        //     return Promise.reject();
        // }

        // return new Promise((resolve, reject) => {
        //     inAppPurchase.getProducts([productId]).then(function (products) {
        //         if (!products || products.length === 0) reject();

        //         inAppPurchase.buy(productId).then(function (data) {
        //             log("buy \n" + JSON.stringify(data));
        //             resolve({
        //                 "productId": productId,
        //                 "date": new Date().toJSON(),
        //                 "signature": data.signature,
        //                 "transactionId": data.transactionId
        //             });
        //         }).catch(function (error) {
        //             reject();
        //             window.logError("APP ERROR purchase " + productId + " failed: " + (error.message || error));
        //         });
        //     }).catch(function (error) {
        //         reject();
        //         window.logError("APP ERROR purchase " + productId + " failed: " + (error.message || error));
        //     });
        // });
    }

    export function promiseMobileServiceProviderCheckDone() {
        if (!providerCheckPromise) {
            providerCheckPromise = new Promise((res, rej) => {
                if (Framework7.device.ios) {
                    res("apple");
                } else {
                    var mobileServiceProviderFallbackTMO = setTimeout(() => {
                        logError("mobileServiceProviderCheck timeout")
                        res("google");
                    }, 5000);

                    if (!window.GooglePlayServicesChecker) {
                        res("google");
                        return;
                    }

                    window.GooglePlayServicesChecker.check(
                        (installedAndUpdated) => {
                            clearTimeout(mobileServiceProviderFallbackTMO)
                            res(installedAndUpdated.status ? "google" : "huawei");
                        },
                        (reason) => {
                            clearTimeout(mobileServiceProviderFallbackTMO)
                            logError("mobileServiceProviderCheck error: " + reason)
                            res("google");
                        }
                    );
                }
            })
        }

        return providerCheckPromise;
    }

    let androidBanners: any = {
        "sl auto": "3951046315",
        "sl auto text": "5026849483",
        "sl cust": "1324882975",
        "sl cust text": "3713767812",
        "sl fix-high text": "3219051599",
        "sl fix-low": "1133311282",
        "sl fix-low text": "4723704955",
        "sl fix-med text": "9784459940",
        "sl fix-high": "6385637962",
        "sl fix-med": "6249906598",
        "sl fix-med fast": "4352184146",
        "sl fix-high text fast": "5282122436",
        "sl fix-med text fast": "6143777965",
        "sl fix-low text fast": "3585897389",
        "sl fix-low fast": "2531496703",
        "sl fix-high fast": "9576590662",
        "sl hms-test2": "1438217947"
    }

    let iosBanners = {
        "sl auto": "8010259566",
        "sl auto text": "2352407071",
        "sl cust": "8940197858",
        "sl cust text": "5409992865",
        "sl fix-high": "8557054470",
        "sl fix-high text": "5493753348",
        "sl fix-low": "7435544490",
        "sl fix-low text": "6806835012",
        "sl fix-med": "4809381158",
        "sl fix-med text": "6639421852",
        "sl fix-high text fast": "9708113296",
        "sl fix-med text fast": "9508307903",
        "sl fix-low fast": "3142704949",
        "sl fix-med fast": "1034404593",
        "sl fix-high fast": "7061591048"
    }


    var curIds;
    var banner;
    export function showAd(show: boolean) {
        promiseDeviceReady().then(() => {
            promiseMobileServiceProviderCheckDone().then(async (provider) => {
                if (!curIds) {
                    let goodAndroidBanners = [
                        androidBanners["sl fix-med"],
                        androidBanners["sl fix-high text"],
                        androidBanners["sl fix-med text"],
                        androidBanners["sl fix-low text"],
                        androidBanners["sl cust text"],
                        androidBanners["sl cust"],
                        androidBanners["sl fix-low"],
                        androidBanners["sl fix-high"],
                        androidBanners["sl fix-med fast"],
                        androidBanners["sl fix-high text fast"],
                        androidBanners["sl fix-med text fast"],
                        androidBanners["sl fix-low text fast"],
                        androidBanners["sl fix-low fast"],
                        androidBanners["sl fix-high fast"]
                    ]
                    if (provider === "huawei") {
                        goodAndroidBanners = [
                            androidBanners["sl hms-test2"]
                        ]
                    }

                    let goodIosBanners = [
                        iosBanners["sl cust text"],
                        iosBanners["sl fix-high text"],
                        iosBanners["sl fix-med text"],
                        iosBanners["sl cust"],
                        iosBanners["sl fix-low"],
                        iosBanners["sl fix-med"],
                        iosBanners["sl fix-high"],
                        iosBanners["sl fix-high text fast"],
                        iosBanners["sl fix-med text fast"],
                        iosBanners["sl fix-low fast"],
                        iosBanners["sl fix-med fast"],
                        iosBanners["sl fix-high fast"]
                    ]

                    curIds = {
                        android: "ca-app-pub-9767837938125687/" + goodAndroidBanners[Math.floor(Math.random() * goodAndroidBanners.length)],
                        ios: "ca-app-pub-9767837938125687/" + goodIosBanners[Math.floor(Math.random() * goodIosBanners.length)]
                    }
                }
                log("showAd " + show, curIds);
                let action = show ? "show" : "hide";
                try {
                    if (!window.admob) {
                        log("window.admob not available")
                        return;
                    }

                    log("start  " + action)
                    if (show) {
                        if (banner) {
                            banner.show()
                        } else {
                            await window.admob.start()
                            banner = new window.admob.BannerAd({
                                adUnitId: curIds[provider === "apple" ? "ios" : "android"]
                            })

                            await banner.show()
                        }

                        // promise = window.admob.banner.show({
                        //     id: curIds
                        // });

                        // nonPersonalizedAds: true,
                        // size: 5,
                        // position: 'bottom'
                    } else {
                        if (banner) {
                            await banner.hide();
                        }
                    }

                } catch (e) {
                    log(e.message);
                    log(e);
                    logError("show ad " + action + " fails " + (e && e.message ? e.message : e));
                }
            });
        })
    }

    export function share(msg, subject?, title?): Promise<any> {
        if (!window.plugins || !window.plugins.socialsharing) return new Promise((res, rej) => { rej() });

        let targets = [
            {
                "title": "e-Mail",
                "id": "email",
                "img": "img/apps/email.png",
                "available": false
            },
            {
                "title": "WhatsApp",
                // this one requires whitelisting of whatsapp:// on iOS9 in your .plist file
                "id": "whatsapp",
                "img": "img/apps/WhatsApp_Logo_2.png",
                "available": false
            },
            {
                "title": "Telegram",
                "id": "org.telegram.messenger",
                "img": "img/apps/Telegram_logo.png",
                "available": false
            },
            {
                "title": "Signal",
                "id": "org.thoughtcrime.securesms",
                "img": "img/apps/signal_logo.png",
                "available": false
            }
        ];

        let targetsCheckPromises = [];
        targets.forEach(target => {
            try {
                let targetCheckPromise = new Promise(function (resolve, reject) {
                    if (target.id === "email") {
                        window.plugins.socialsharing.canShareViaEmail(function (e) { target.available = true; resolve(target) }, function (e) { resolve(target) })
                    } else {
                        window.plugins.socialsharing.canShareVia(target.id, msg, null, null, null, function (e) { target.available = true; resolve(target) }, function (e) { resolve(target) });
                    }
                });

                targetsCheckPromises.push(targetCheckPromise);
            } catch (e) {
                logError("check share taget fails ", e);
            }
        })

        return new Promise(function (resolveShare, rejectShare) {
            let buttons = [];
            Promise.all(targetsCheckPromises).then(targets => {
                let availableTargets = targets.filter(target => target.available);

                availableTargets.push({
                    "title": $translate.instant("more"),
                    "img": "img/apps/more.png",
                    "available": true
                })

                availableTargets.forEach((target, idx) => {
                    let targetGroupIdx = Math.floor(idx / 3);

                    if (!buttons[targetGroupIdx]) buttons[targetGroupIdx] = [];

                    buttons[targetGroupIdx].push({
                        text: target.title,
                        icon: '<img src="' + target.img + '" width="48"/>',
                        onClick: () => {
                            try {
                                if (target.id === "email") {
                                    if (cordova.plugins.email) {
                                        log("cordova.plugins.email ", subject, msg);
                                        cordova.plugins.email.open(
                                            {
                                                "subject": subject,
                                                "body": msg,
                                                "isHtml": false
                                            },
                                            () => { resolveShare(); },
                                            (e) => { log("share via " + target.id + " failed!", e); rejectShare(); }
                                        );
                                    } else {
                                        log("fallback to " + window.plugins.socialsharing.shareViaEmail)

                                        window.plugins.socialsharing.shareViaEmail(
                                            msg,
                                            subject,
                                            null,
                                            null,
                                            null,
                                            null,
                                            null,
                                            () => { resolveShare() },
                                            (e) => { rejectShare(); log("share via " + target.id + " failed!", e) }
                                        )
                                    }
                                } else if (target.id) {
                                    if (subject) msg = subject + "\n\n" + msg

                                    if (target.id === "whatsapp") {
                                        window.location.href = 'whatsapp://send?text=' + encodeURIComponent(msg);
                                        resolveShare()
                                        return;
                                    }

                                    window.plugins.socialsharing.shareVia(
                                        target.id,
                                        msg,
                                        null,
                                        null,
                                        null,
                                        () => { resolveShare() },
                                        (e) => { rejectShare(); log("share via " + target.id + " failed!", e) }
                                    );
                                } else {
                                    if (subject) msg = subject + "\n\n" + msg

                                    window.plugins.socialsharing.share(
                                        msg,
                                        null,
                                        null,
                                        null,
                                        () => { resolveShare() },
                                        (e) => { rejectShare(); log("share failed!", e) }
                                    );
                                }
                            } catch (e) {
                                rejectShare();
                                log("share via " + target.id + " failed!", e)
                            }
                        }
                    });
                });

                if (title) {
                    buttons.unshift([
                        {
                            text: title,
                            label: true
                        }
                    ]);
                }

                let shareActionSheet = vm.$f7.actions.create({
                    grid: true,
                    buttons: buttons
                });
                shareActionSheet.open();
            })
        });
    }

    export function writeFile(fileName: string, data: string, tryLocalStorage: boolean) {
        return new Promise(function (resolve, reject) {
            if (tryLocalStorage || !window.requestFileSystem) {
                try {
                    window.localStorage.setItem("_file_" + fileName, data);
                } catch (e) {
                    log("try localStorage failed " + e.message);
                }
            }

            promiseDeviceReady().then(function () {
                if (!window.requestFileSystem) {
                    log("requestFileSystem not available")
                    reject();
                    return;
                }

                window.requestFileSystem(LocalFileSystem.PERSISTENT, 0, function (fs) {
                    log('file system open: ' + fs.name + " " + cordova.file.dataDirectory + fileName + "--" + fs.root.toURL());
                    fs.root.getFile(fileName, { create: true, exclusive: false }, function (fileEntry) {
                        log("fileentry:" + fileEntry.toURL())
                        // fileEntry.name == 'someFile.txt'
                        // fileEntry.fullPath == '/someFile.txt'
                        fileEntry.createWriter(function (fileWriter) {
                            fileWriter.onwriteend = function () {
                                log("Successful file write...");
                                resolve();
                            };

                            fileWriter.onerror = function (e) {
                                log("Failed file write: " + e.toString());
                                reject(e);
                            };

                            fileWriter.write(data);
                        });
                    }, reject);
                }, reject);
            });
        });
    }


    export function readFile(fileName: string, tryLocalStorage: boolean) {
        return new Promise(function (resolve, reject) {
            if (tryLocalStorage) {
                try {
                    let lsData = window.localStorage.getItem("_file_" + fileName);

                    if (lsData) {
                        log("Successful localstorage read")
                        resolve(lsData);
                        return;
                    }
                } catch (e) {
                    log("try read from localStorage failed " + e.message);
                }
            }

            promiseDeviceReady().then(
                function () {
                    if (!window.requestFileSystem) {
                        log("requestFileSystem not available")
                        reject();
                        return;
                    }

                    window.requestFileSystem(LocalFileSystem.PERSISTENT, 0, function (fs) {
                        log('file system open: ' + fs.name + " " + cordova.file.dataDirectory + fileName + "--" + fs.root.toURL());
                        fs.root.getFile(fileName, { create: true, exclusive: false }, function (fileEntry) {
                            log("fileentry:" + fileEntry.toURL())

                            fileEntry.file(function (file) {
                                let reader = new FileReader();

                                reader.onloadend = function () {
                                    log("Successful file read: " + fileEntry.fullPath);
                                    resolve(this.result);
                                };

                                reader.readAsText(file);

                            }, reject);
                        }, reject);
                    });
                });
        });
    }

    export function numberWordToNumber(w) {
        if (!vm.wordsToNumber || !vm.wordsToNumber.length) return 0;

        if (!w) return 0;
        w = w.toLowerCase();

        let erg = 0;
        //TODO sort all numwords by word length to one array?
        for (let m = vm.wordsToNumber.length; m--;) {
            var map = vm.wordsToNumber[m];
            for (let numWord in map) {
                let oldLength = w.length;
                w = w.replace(numWord, "");

                if (oldLength !== w.length) {
                    erg += map[numWord];
                    break;
                }
            }
        };

        var wRest = w;
        //remove number bind words (!after replacement of all known numberwords)
        if (vm.numberJoinWordRegExp) wRest = w.replace(vm.numberJoinWordRegExp, "");

        return wRest.length === 0 ? erg : 0;
    }

    export let andSplitWords = {
        en: / and |^and | and$/,
        mg: / ary |^ary | ary$/,
        sm: / ma |^ma | ma$/,
        ro: / și |^și | și$/,
        ru: / и |^и | и$/,
        sr: / и |^и | и$/,
        sw: / na |^na | na$/,
        fy: / en |^en | en$/,
        tl: / at |^at | at$/,
        cy: / a |^a | a$/,
        vi: / và |^và | và$/,
        yi: /און/,
        af: / en |^en | en$/,
        ar: /و/,
        sq: / dhe |^dhe | dhe$/,
        bs: / i |^i | i$/,
        bn: / আনড |^আনড | আনড$/,
        bg: / и |^и | и$/,
        my: /နှင့်/,
        hr: / i |^i | i$/,
        co: / è |^è | è$/,
        ca: / i |^i | i$/,
        ny: / ndi |^ndi | ndi$/,
        cs: / a |^a | a$/,
        da: / og |^og | og$/,
        az: / və |^və | və$/,
        fr: / et |^et | et$/,
        eu: / eta |^eta | eta$/,
        ml: /ഒപ്പം/,
        ne: / अण्ड |^अण्ड | अण्ड$/,
        st: / le |^le | le$/,
        sd: /۽/,
        sn: / and |^and | and$/,
        su: / sareng |^sareng | sareng$/,
        tg: / ва |^ва | ва$/,
        ta: /மற்றும்/,
        si: /සහ/,
        uk: / і |^i | i$/,
        tr: / ve |^ve | ve$/,
        uz: / va |^va | va$/,
        xh: / kwaye |^kwaye | kwaye$/,
        ur: /اور/,
        zu: / futhi |^futhi | futhi$/,
        be: / і |^і | і$/,
        hy: / եւ |^եւ | եւ$/,
        fi: / ja |^ja | ja$/,
        et: / ja |^ja | ja$/,
        el: / και |^και | και$/,
        de: / und |^und | und$/,
        lv: / un |^un | un$/,
        ko: /과 |와 /,
        ky: / жана |^жана | жана$/,
        kn: /ಮತ್ತ/,
        kk: / және |^және | және$/,
        it: / e |^e | e$/,
        ga: / agus |^agus | agus$/,
        es: / y | e |^e |^y | e$| y$/,
        so: / iyo |^iyo | iyo$/,
        sl: / in |^in | in$/,
        ms: / dan |^dan | dan$/,
        mi: / a |^a | a$/,
        mt: / u |^u | u$/,
        no: / og |^og | og$/,
        mn: /болон/,
        pl: / i |^i | i$/,
        ku: / û |^û | û$/,
        la: / et |^et | et$/,
        km: /និង/,
        ja: /と/,
        ig: / ma |^ma | ma$/,
        is: / og |^og | og$/,
        hu: / és |^és | és$/,
        th: /และ/,
        sk: / a |^a | a$/,
        lo: /ແລະ/,
        yo: / ati |^ati | ati$/,
        mk: / и |^и | и$/,
        ka: /და/,
        hi: /तथा/,
        lb: / an |^ | $/,
        te: /మరియు/,
        lt: / ir |^ir | ir$/,
        nl: / en |^en | en$/,
        eo: / kaj |^kaj | kaj$/,
        gd: / agus |^agus | agus$/,
        gl: / e |^e | e$/,
        mr: /आणि/,
        ps: /و/,
        pt: / e |^e | e$/,
        sv: / och |^och | och$/,
        ha: /da kuma/,
        ht: / ak |^ak | ak$/,
        gu: /અને/,
        fa: /و/
    }


    // export function prepareRecordSpeech() {
    //     return new Promise((resolve, reject) => {
    //         promiseDeviceReady().then(() => {
    //             let permissions = window.plugins && window.plugins.permissions;
    //             if (!permissions) {
    //                 permissions = { requestPermission: (perm, suc, err) => { suc({ "hasPermission": true, "dummy": true }) } };
    //             }

    //             if (SpeechRecognition) {
    //                 permissions.requestPermission(
    //                     permissions.RECORD_AUDIO,
    //                     (status) => {
    //                         log("hasPermission ", status);
    //                         try {
    //                             if (status.hasPermission) {
    //                                 //TODO try window.plugins.speechRecognition.startListening instead of new SpeechRecognition()
    //                                 resolve(new SpeechRecognition());
    //                             } else {
    //                                 log("permission RECORD_AUDIO denied");
    //                                 vm.$f7.toast.create({
    //                                     text: $translate.instant("mic_permission_req"),
    //                                     closeTimeout: 8000,
    //                                     destroyOnClose: true
    //                                 }).open();
    //                                 reject();
    //                             }
    //                         } catch (e) {
    //                             log("new SpeechRecognition failed ", e);
    //                             reject(e);
    //                         }
    //                     },
    //                     () => {
    //                         reject();
    //                     }
    //                 );
    //             } else {
    //                 log("SpeechRecognition not available");
    //                 reject();
    //             }
    //         }, reject);
    //     });
    // }


    export function recordSpeech(lang, onText, onError, msg?) {
        window.plugins.speechRecognition.requestPermission(
            () => {
                let options = {
                    language: lang,
                    matches: 1,
                    prompt: msg || "",      // Android only
                    showPopup: !!msg,  // Android only
                    showPartial: true
                }

                window.plugins.speechRecognition.startListening(
                    (matches) => { onText(matches[0]) },
                    (err) => { onError(err) },
                    options
                )
            },
            () => {
                log("permission RECORD_AUDIO denied");
                vm.$f7.toast.create({
                    text: $translate.instant("mic_permission_req"),
                    closeTimeout: 7000,
                    destroyOnClose: true,
                    closeButton: true,
                    closeButtonText: $translate.instant("ok")
                }).open();
                onError("permission RECORD_AUDIO denied");
            }
        );
    }

    export function stopRecordSpeech() {
        return new Promise((resolve, reject) => {
            promiseDeviceReady().then(() => {
                console.log("stop");
                window.plugins.speechRecognition.stopListening(
                    () => { console.log("stopped"); resolve("ok") },
                    (err) => { reject(err) }
                )
            })
        });
    }

    export function toJSON(object) {
        return JSON.stringify(object, function (key, val) { if (key !== "tmp") return val; });
    }

    export function requestJSON(url) {
        return new Promise(function (resolve, reject) {
            Framework7.request.json(
                url,
                function (data, status, xhr) {
                    log(url + " > " + status);
                    if ((status + "").charAt(0) === "2" || status === 0) {
                        resolve(data, status, xhr);
                    } else {
                        reject(status, xhr)
                    }
                }, function (xhr, status) {
                    log(url + " !> " + status + "  " + xhr.statusText);
                    reject(status, xhr);
                }
            );
        });
    }

    export function request(url, method?, data?, options?) {
        return new Promise(function (resolve, reject) {
            if (!options) options = {};

            options.url = url;
            options.method = method || "GET";
            options.data = data;

            if (!options.cache) options.cache = false;
            if (!options.contentType) options.contentType = "application/json";
            if (!options.processData) options.processData = true;
            if (!options.dataType) options.dataType = "json";
            if (!options.timeout) options.timeout = options.timeout === void (0) ? 20000 : options.timeout;
            options.success = function (data, status, xhr) {
                if ((status + "").charAt(0) === "2" || (method === "DELETE" && status === 404)) {
                    resolve(data, xhr.status, xhr);
                } else {
                    reject(xhr.status, xhr)
                }
            }
            options.error = function (xhr) {
                reject(xhr.status, xhr);
            }

            Framework7.request(options);
        });
    }

    export function requestBlob(url, method?, data?, options?) {
        if (!options) options = {};

        return new Promise((resolve, reject) => {
            let xhr = new XMLHttpRequest();
            xhr.open(method || "GET", url, true);
            xhr.responseType = "blob";

            let xhrTimeout;
            xhr.onload = (oEvent) => {
                if (xhrTimeout) clearTimeout(xhrTimeout);

                if ((xhr.status + "").charAt(0) === "2" || (method === "DELETE" && xhr.status === 404)) {
                    resolve(xhr.response, xhr.status, xhr);
                } else {
                    reject(xhr.status, xhr)
                }
            };
            xhr.onerror = (xhr, status) => {
                log("onerror")
                if (xhrTimeout) clearTimeout(xhrTimeout);
                reject(xhr.status, xhr);

            };
            xhr.onabort = (xhr, status) => {
                if (xhrTimeout) clearTimeout(xhrTimeout);
                reject(xhr.status, xhr);
            };
            xhr.ontimeout = () => {
                if (xhrTimeout) clearTimeout(xhrTimeout);
                reject(xhr.status, xhr);
            };

            xhrTimeout = setTimeout(function () {
                xhr.abort();
            }, options.timeout || 20000);

            xhr.send(data);
        });
    }

    export function id(mask, map) {
        return vm.$f7.utils.id(mask, map);
    }

    let normalizedStringCache = {};
    export function normalizeStr(str) {
        if (!str) return "";

        // let hit = normalizedStringCache[str];
        // if (hit) return hit;
        // return normalizedStringCache[str] = Framework7.utils.removeDiacritics(str).toLowerCase();
        return Framework7.utils.removeDiacritics(str).toLowerCase();
    }

    export function indexOfItem(itemName: string, inArray: Item[]) {
        if (!itemName) return -1;

        for (let i = 0; i < inArray.length; i++) {
            const curItem = inArray[i];
            if (!curItem) continue;

            if (curItem.name === itemName || thatlist.util.normalizeStr(curItem.name) === thatlist.util.normalizeStr(itemName)) {
                return i;
            }
        }

        return -1;
    }

    export function getUnit(str) {
        let normStr = normalizeStr(str);

        for (let u = 0; u < app.tmp.units.length; u++) {
            let curUnit = app.tmp.units[u];
            if (curUnit.normTitle == normStr) return curUnit;
        }
    }
    // private wunderlistItemRxp = /(❏ |❏)/gm;
    // private wunderlistPickedItemRxp = /(✔ |✔)/gm;
    export function parseItemsText(text) {
        let items = [];
        try {
            if (!text || text.length > 10000) return;

            //clear blank lines
            text = text.replace(/^\s*[\r\n]/gm, "");

            let lines = text.split("\n");

            let catsByTitle = {};
            for (let catName in app.categories) {
                catsByTitle[app.categories[catName].title] = catName;
            }

            let curCat, isAppData;
            for (let l = 0; l < lines.length; l++) {
                let curLine = lines[l];
                console.log(curLine);
                curLine = curLine.trim();

                //if is cat
                if (curLine.charAt(0) == "[") {
                    isAppData = true;
                    curCat = catsByTitle[curLine.substring(1, curLine.indexOf("]"))]
                    continue;
                }

                if (curLine.length < 2 || curLine.charAt(curLine.length - 1) == ":") continue;

                var trustData = isAppData || lines.length === 1;

                if (curLine.length > (trustData ? 500 : 60)) {
                    let curLineParts = curLine.split(",");
                    if (curLineParts.length > 3 && lines.length < 5) {
                        //expect goods are , separeted
                        let args = [l + 1, 0].concat(curLineParts);
                        //add comma separated parts as lines
                        Array.prototype.splice.apply(lines, args);
                    } else {
                        //ignore text line
                    }

                    continue;
                }

                try {
                    curLine = curLine.replace("\u00bc", "1/4").replace("\u00bd", "1/2").replace("\u00be", "3/4");

                    let matchOrig = curLine.match(fracRegExp);
                    if (matchOrig) {
                        let match = matchOrig[0].split(" ");
                        let m = 0;
                        while (match.length) {
                            m += eval(match.shift());
                        }
                        curLine = curLine.replace(matchOrig[0], m);
                    }
                } catch (e) {
                    window.logError("APP ERROR paste parse: " + e.message + (e.stack ? "\n" + e.stack : ""))
                }

                let item = thatlist.util.parseItemName(curLine, !trustData);
                if (item) {
                    item.cat = curCat;
                    items.push(item)
                }
            }
        }
        catch (e) {
            window.logError("APP ERROR paste: " + e.message + (e.stack ? "\n" + e.stack : ""))
        }

        return items;
    }

    export function parseItemName(name: string, clean?: boolean) {
        if (!name) return;

        name = name.replace(replWSRegExp, " ").replace(replBSRegExp, "");

        if (clean) {
            //normalize e.g. (n) /n (en) (er)
            name = name.replace(replBracketsRegExp, "$1").replace(replExplRegExp, "").replace(replSlashNRegExp, "$1").replace(replSlashSRegExp, "$1");
        }

        name = name.trim();

        let parsedItem;
        if (!app.settings.extract_count) {
            parsedItem = {
                "name": name
            }
        } else {
            //support Hack 1,5Kg, halb und halb
            let descSepIdx = name.lastIndexOf(",");

            let desc, parts;
            //support Hack 1,5Kg
            if (descSepIdx >= 0 && isNaN(Number(name.charAt(descSepIdx - 1)))) {
                //if desc is only one word it is expected to be specialisation e.g. Vine, white
                desc = name.substring(descSepIdx + 1).trim();
                parts = name.substring(0, descSepIdx).split(" ");
            } else {
                parts = name.split(" ");
            }

            let shortDesc = desc ? desc.split(app.getLanguage() === "ja" ? "、" : ",")[0].trim() : "";
            if (clean && shortDesc.indexOf(" ") > 0) {
                shortDesc = "";
            }

            for (let p = 0; p < parts.length; p++) {
                let curPart = parts[p];

                let count;
                //count starts with number e.g. 2g 

                //curPart starts with number
                if (!isNaN(curPart.charAt(0)) && (p === 0 || p === parts.length - 1)) {
                    let count = curPart;
                    let countReplaced = count.replace(",", ".")


                    //curPart is number
                    if (!isNaN(countReplaced)) {
                        let nextPart = parts[p + 1];
                        //check if next part is unit
                        let unit;
                        if (nextPart && (parts[p + 2] || parts[p - 1]) && (nextPart.length <= 2 || (unit = getUnit(nextPart)))) {
                            count += unit ? count == 1 ? unit.title_sing : unit.title : nextPart;

                            parts.splice(p, 2);
                        } else {
                            parts.splice(p, 1);
                            count += "";
                        }
                    } else {
                        parts.splice(p, 1);
                    }

                    let isSing = count == "1" || count.charAt(0) == "1" && (count.charAt(1) == " " || isNaN(count.charAt(1)))

                    parsedItem = {
                        "isSing": isSing,
                        "count": count,
                        "unit": count.replace(/[0-9,.]/g, '') + (count.charAt(count.length - 1) == "." ? "." : ""),
                        "name": parts.join(" ") + (shortDesc ? ", " + shortDesc : "")
                    }

                    break;
                }
            }

            if (!parsedItem) {
                parsedItem = {
                    "name": parts.join(" ") + (shortDesc ? ", " + shortDesc : "")
                }
            }
        }

        parsedItem.name = parsedItem.name.charAt(0).toUpperCase() + parsedItem.name.substring(1);

        return parsedItem;
    }

    export function scrollCurrentPageTo(scrollPos, add?: boolean): number {
        let $elem = $$(vm.$f7.views.main.router.currentPageEl).find(".page-current .page-content");
        let elem = $elem[0];
        if (!elem) return;

        let oldScrollPos = elem.scrollTop;
        $elem.scrollTop(0);
        elem.scrollTop = scrollPos + (add ? oldScrollPos : 0);
        return oldScrollPos;
    }

    export function focusInput(sel: string, selectText?: boolean) {
        if (window.event) {
            window.event.stopPropagation();
            window.event.preventDefault();
        }
        let elem = $$(sel).focus()[0];
        if (selectText) elem.setSelectionRange(0, 1000);
        thatlist.util.showKeyboard();
        // Vue.nextTick(() => {
        //     let elem = $$(sel).focus()[0];

        //     if (elem) {
        //         elem.blur();
        //         elem.focus();
        //         if (selectText) elem.setSelectionRange(0, 1000);
        //         thatlist.util.showKeyboard();
        //     }
        //     let focusInt = window.setInterval(function () {
        //         let elem = $$(sel).focus()[0];
        //         if (elem) {
        //             window.clearInterval(focusInt);
        //             //elem.blur();
        //             elem.focus();
        //             if (selectText) elem.setSelectionRange(0, 1000);
        //             thatlist.util.showKeyboard();
        //         }
        //     }, 20);
        //     window.setTimeout(function () {
        //         window.clearInterval(focusInt)
        //     }, 160);
        // });
    }

    /**
     * 
     * @param cc current count
     * @param nc new count
     */
    export function getEstimatedCountSums(cc, nc) {
        let altCount1: any = "";
        let altCount2: any = "";

        if (!cc && !nc) {
            console.log("a");
            altCount1 = "2";
        }

        //has cur count unit
        let hasCcu: any = cc && isNaN(cc);
        let ccu = "";
        //cur count value
        let ccv;
        if (hasCcu) {
            let ccvRange = thatlist.util.numberRangeOfString(cc);
            //if contains number
            if (ccvRange[1]) {
                let numVal = cc.substring(ccvRange[0], ccvRange[1]).replace(",", ".");
                console.log(cc, numVal);
                ccv = parseFloat(numVal);
                //skip unit for unespected format (count does not start with a number)
                ccu = ccvRange[0] === 0 ? cc.substring(ccvRange[1]) : "";
            }
        } else {
            ccv = parseFloat(((cc || 1) + "").replace(",", "."));
        }

        let hasNcu: any = nc && isNaN(nc);
        let ncu = "";
        let ncv;
        if (hasNcu) {
            let ncvRange = thatlist.util.numberRangeOfString(nc);
            //if contains number
            if (ncvRange[1]) {
                let numVal = nc.substring(ncvRange[0], ncvRange[1]).replace(",", ".");
                ncv = parseFloat(numVal);
                //skip unit for unespected format (count does not start with a number)
                ncu = ncvRange[0] === 0 ? nc.substring(ncvRange[1]) : "";
            }
        } else {
            ncv = parseFloat(((nc || 1) + "").replace(",", "."));
        }

        if (!cc && nc) {
            console.log("b " + ncv);
            if (!ncv) {
                altCount1 = nc;
            } else {
                altCount1 = (ncv + 1) + ncu;
                altCount2 = (ncv * 2) + ncu;
            }
        } else if (cc && !nc) {
            console.log("c " + ccv);
            if (!ccv) {
                altCount1 = cc;
            } else {
                altCount1 = (ccv + 1) + ccu;
                altCount2 = (ccv * 2) + ccu;
            }
        } else if (cc && nc) {
            if (ccu && ncu) {
                console.log(ccu, ncu);
                if (ccu.trim().toLowerCase() === ncu.trim().toLowerCase()) {
                    console.log("d");
                    altCount1 = (ccv + ncv) + ccu;
                } else {
                    console.log("e");
                    altCount1 = (ncv + 1) + ncu;
                    altCount2 = (ncv * 2) + ncu;
                }
            } else if (!ccu && !ncu) {
                altCount1 = (ccv + ncv);
                console.log("f");
                if (!cc || !nc) altCount2 = (!cc ? ncv : ccv) * 2;
            } else {
                console.log("g");
                altCount1 = (ccv + ncv) + ncu;
                altCount2 = (ncv * 2) + ncu;
            }
        }

        var alts = [altCount1];

        if (altCount2 && altCount2 !== altCount1) alts.push(altCount2);

        return alts;
    }

    export function numberRangeOfString(str) {
        let start, end;
        for (let c = 0; c < str.length; c++) {
            let char = str.charAt(c);
            let charIsNumber = !isNaN(parseInt(char, 10));
            if (start === void (0)) {
                if (charIsNumber) start = c;
            } else if (!charIsNumber) {
                //handle decimal numbers
                if (char === "." || char === ",") continue;
                end = c;
                break;
            }
        }

        return [start, end]
    }

    export function onFocusCount(elem) {
        if (Framework7.device.ios) {
            var scrollTop = document.body.scrollTop;
            window.scrollTo(0, 0);
            document.body.scrollTop = 0;
            let pageContent = vm.$f7.$(".view-main .page-current .page-content")[0];
            if (pageContent) pageContent.style.marginTop = -scrollTop + "px";
            // Keyboard.disableScroll(true);
        }

        //TODO mark first numbers only
        var numberRange = numberRangeOfString(elem.value);

        if (numberRange[1]) elem.setSelectionRange(numberRange[0], numberRange[1]);
        else elem.setSelectionRange(0, elem.value.length);
    }

    export function onBlurCount(elem) {
        if (Framework7.device.ios) {
            vm.$f7.$(".view-main .page-current .page-content")[0].style.marginTop = "";
            // Keyboard.disableScroll(false);
        }
    }

    /**
     * from https://github.com/angular-translate/angular-translate/blob/master/src/service/translate.js Line 93
     */
    export function getFirstBrowserLanguage(): string {
        let nav = navigator,
            browserLanguagePropertyKeys = ['language', 'browserLanguage', 'systemLanguage', 'userLanguage'],
            i,
            language;

        // support for HTML 5.1 "navigator.languages"
        var lang = "en";

        if (Array.isArray(nav.languages)) {
            for (i = 0; i < nav.languages.length; i++) {
                language = nav.languages[i];
                if (language && language.length) {
                    lang = language.substring(0, 2).toLowerCase();
                    break;
                }
            }
        }

        // support for other well known properties in browsers
        for (i = 0; i < browserLanguagePropertyKeys.length; i++) {
            language = nav[browserLanguagePropertyKeys[i]];
            if (language && language.length) {
                lang = language.substring(0, 2).toLowerCase();
                break;
            }
        }

        var langMap = {
            "nb": "no",
            "nn": "no"
        }

        return langMap[lang] || lang;
    };
}